"""
See [here](https://refactoring.guru/design-patterns/composite).

Examples:
    >>> import rna
    >>> from rna.pattern.composite import Component, Leaf, Composite
    >>> class MyLeaf(Leaf):
    ...     def operation(self) -> str:
    ...         return "Leaf"

    >>> class MyComposite(Composite):
    ...     def operation(self) -> str:
    ...         results = []
    ...         for child in self._children:
    ...             results.append(child.operation())
    ...         return f"Branch({'+'.join(results)})"

    The client code works with all of the components via the base interface.
    >>> def client_code(component: Component) -> None:
    ...     print(f"RESULT: {component.operation()}", end="")

    Thanks to the fact that the child-management operations are declared in the
    base Component class, the client code can work with any component, simple or
    complex, without depending on their concrete classes.
    >>> def client_code2(component1: Component, component2: Component) -> None:
    ...     if component1.is_composite():
    ...         component1.add(component2)
    ...     print(f"RESULT: {component1.operation()}", end="")


    This way the client code can support the simple leaf components...
    >>> simple = MyLeaf()

    Client: I've got a simple component:
    >>> client_code(simple)
    RESULT: Leaf

    ...as well as the complex composites.
    >>> tree = MyComposite()

    >>> branch1 = MyComposite()
    >>> branch1.add(MyLeaf())
    >>> branch1.add(MyLeaf())

    >>> branch2 = MyComposite()
    >>> branch2.add(MyLeaf())

    >>> tree.add(branch1)
    >>> tree.add(branch2)

    Client: Now I've got a composite tree:
    >>> client_code(tree)
    RESULT: Branch(Branch(Leaf+Leaf)+Branch(Leaf))

    Client: I don't need to check the components classes even when managing the tree:
    >>> client_code2(tree, simple)
    RESULT: Branch(Branch(Leaf+Leaf)+Branch(Leaf)+Leaf)

    This pattern is generically implemented such that you can subclass Component and then
    auto-generate Leave and Composite classes from this.
    >>> from abc import abstractmethod
    >>> class Graphic(Component):
    ...     @abstractmethod
    ...     def render(self) -> None:
    ...         pass
    >>> class CompositeGraphic(Graphic.Composite):
    ...     def render(self):
    ...         for child in self._children:
    ...             child.render()
    >>> class Ellipse(Graphic.Leaf):
    ...     def __init__(self, name):
    ...         self.name = name
    ...
    ...     def render(self):
    ...         print("Ellipse", self.name)
    >>> ellipse1 = Ellipse("1")
    >>> ellipse2 = Ellipse("2")
    >>> ellipse3 = Ellipse("3")
    >>> ellipse4 = Ellipse("4")
    >>> graphic1 = CompositeGraphic()
    >>> graphic2 = CompositeGraphic()
    >>> graphic1.add(ellipse1)
    >>> graphic1.add(ellipse2)
    >>> graphic1.add(ellipse3)
    >>> graphic2.add(ellipse4)
    >>> graphic = CompositeGraphic()
    >>> graphic.add(graphic1)
    >>> graphic.add(graphic2)
    >>> graphic.render()
    Ellipse 1
    Ellipse 2
    Ellipse 3
    Ellipse 4

"""
import typing
from abc import ABCMeta, abstractmethod


class ComponentMeta(ABCMeta):
    """
    Meta class exposing class attributes as mixin generators
    """

    # pylint: disable=invalid-name
    def Subclass(cls, name: str, mixins: typing.Type, namespace: dict = None):
        """
        Auto-generate a sub class from this class together with mixins
        Args:
            name: name of new class
        """
        if namespace is None:
            namespace = {}
        return type(name, (cls,) + mixins, namespace)

    @property
    def Leaf(cls) -> typing.Type:  # pylint: disable=invalid-name
        """
        Auto-generate a Leaf mixin from this class.
        The Leaf class represents the end objects of a composition. A leaf can't
        have any children.  Usually, it's the Leaf objects that do the actual work, whereas
        Composite objects only delegate to their sub-components.
        """

        def add(self, component: cls):  # pylint: disable=no-self-use
            """
            Leafes can not add
            """
            raise TypeError(
                "Trying to add component {component} to Leaf".format(**locals())
            )

        def remove(self, component: cls):  # pylint: disable=no-self-use
            """
            Leafes can not remove
            """
            raise TypeError(
                "Trying to remove component {component} from Leaf".format(**locals())
            )

        return cls.Subclass("Leaf", (), dict(add=add, remove=remove))

    @property
    def Composite(cls) -> typing.Type:  # pylint: disable=invalid-name
        """
        Auto-generate a Composite subclass from this class.
        The Composite class represents the complex components that may have children. Usually, the
        Composite objects delegate the actual work to their children and then "sums-up" the result.
        """

        def __init__(self) -> None:
            self._children: typing.List[cls] = []

        def add(self, component: cls) -> None:
            """
            Add component by appending it to children
            """
            self._children.append(component)  # pylint:disable=protected-access
            component.parent = self

        def remove(self, component: cls) -> None:
            """
            Remove component by appending it to children
            """
            self._children.remove(component)  # pylint:disable=protected-access
            component.parent = None

        def is_composite(self) -> bool:  # pylint: disable=unused-argument
            """
            Whether a component can bear children.
            """
            return True

        return cls.Subclass(
            "Composite",
            (),
            dict(__init__=__init__, add=add, remove=remove, is_composite=is_composite),
        )


class Component(metaclass=ComponentMeta):
    """
    The base Component class declares common operations for both simple and complex objects of a
    composition.
    """

    @property
    def parent(self) -> "Component":
        """
        Returns:
            parent component
        """
        return self._parent

    @parent.setter
    def parent(self, parent: "Component"):
        """
        Optionally, the base Component can declare an interface for setting and
        accessing a parent of the component in a tree structure. It can also
        provide some default implementation for these methods.
        """
        self._parent = parent

    # In some cases, it would be beneficial to define the child-management
    # operations right in the base Component class. This way, you won't need to
    # expose any concrete component classes to the client code, even during the
    # object tree assembly.

    @abstractmethod
    def add(self, component: "Component") -> None:
        """
        Polymorphism for adding components to a subclass.
        """

    @abstractmethod
    def remove(self, component: "Component") -> None:
        """
        Polymorphism for removing components to a subclass.
        """

    def is_composite(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether a component can bear children.
        """
        return False


class Leaf(Component.Leaf):
    """
    Implementation for pattern namespace. If you want to subclass Component, use auto-generated
    Component.Leaf type
    """


class Composite(Component.Composite):
    """
    Implementation for pattern namespace. If you want to subclass Component, use auto-generated
    Component.Composite type
    """



"""
process utilities
"""
import typing
import shlex
import subprocess
import logging


def execute(
    system_command: str,
    level: typing.Optional[int] = None,
    logger: typing.Optional[typing.Union[logging.Logger, bool]] = None,
    output: bool = False,
    **kwargs
) -> typing.Optional[typing.List[str]]:
    """
    Execute a system command, passing STDOUT and STDERR to logger.

    Args:
        system_command: code to execute
        level: logging level of the message (not of the logger). Defaults to DEBUG
        logger: if not None, log the command an also automaticaly infer output = True.
            if logger==True create a logger from the name
        output: if True, return the output. Is False by default because communicate can
            eventually overflow the memory buffer. If False returns None
        **kwargs: forwarded to Popen

    Examples:
        >>> import rna
        >>> import logging
        >>> log = logging.getLogger(__name__)

        Returns the output if ....

        ... any logging is involved
        >>> rna.process.execute("echo 'Test'", level=logging.ERROR)
        ['Test']
        >>> rna.process.execute("echo 'Test'", logger=log)
        ['Test']

        ... or explicitly asked for
        >>> rna.process.execute("echo 'Test'", level=None, output=True)
        ['Test']

        To be more efficient
        >>> rna.process.execute("echo 'Test'", level=None)

    """
    if level is not None and logger is None:
        logger = True
    if logger and isinstance(logger, bool):  # i.e. True
        logger = logging.Logger("Output of {}".format(system_command))
    if logger or output:
        if level is None:
            level = logging.DEBUG
        kwargs.setdefault("text", True)
        kwargs.setdefault("stdout", subprocess.PIPE)
        kwargs.setdefault("stderr", subprocess.STDOUT)
    if logger:
        logger.log(level, "Executing system command: '%s'", system_command)
    pipe = subprocess.Popen(shlex.split(system_command), **kwargs)
    if logger:
        out = []
        for line in iter(pipe.stdout.readline, ""):
            line = line.strip("\n")
            logging.log(level, line)
            out.append(line)
        pipe.stdout.close()
        return_code = pipe.wait()
    elif output:
        out, return_code = pipe.communicate()
        out = out.rstrip("\n").split("\n")
    else:
        out = None
        return_code = pipe.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, system_command)
    return out



"""
Matplotlib implementation of an interchangeable backend for plotting
"""
import typing
import logging
import os
import contextlib
import re
import functools
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.animation as manimation
import matplotlib.dates as dates
import matplotlib.patches as mplpatch
import mpl_toolkits.mplot3d as plt3d
import mpl_toolkits.axes_grid1 as pltg1
import rna
import numpy as np


# ANIMATION

try:
    FFMPEGWRITER = manimation.writers["ffmpeg"]
    FFMPEG = True
except Exception as err:
    logging.getLogger().warning(str(err))
    FFMPEGWRITER = object
    FFMPEG = False


class Writer(FFMPEGWRITER):
    """
    FFMPEGWRITER child

    Args:
        **kwargs:
            title (str)
            comment (str)
            fps (int)

    Examples:
        >>> import numpy as np
        >>> import rna
        >>> from rna.plotting.backends.matplotlib import Writer
        >>> axes = rna.plotting.gca()
        >>> writer = Writer()
        >>> pictures = np.array([[[1,1], [0,0]],
        ...                      [[0.5,1], [0,0.5]]])
        >>> with writer.saving():
        ...     for s in pictures:
        ...         artist = axes.imshow(s.T)
        ...         writer.grab_frame()
        ...         axes.clear()
    """

    def __init__(self, **kwargs):
        if FFMPEG:
            # set Defaults
            title = kwargs.pop("title", "")
            comment = kwargs.pop("comment", "")
            artist = kwargs.pop("artist", "rna plotting module")
            fps = kwargs.pop("fps", 15)
            if len(kwargs) > 0:
                raise ValueError("Some Attributes are not used.")

            # set up writer
            metadata = dict(title=title, artist=artist, comment=comment)
            super(Writer, self).__init__(fps=fps, metadata=metadata)

    @contextlib.contextmanager
    def saving(self, fig=None, path=None, dpi=300, **kwargs):
        """
        Overwritinig to FFMPEGWRITER.saving
        setting defaults though
        """
        log = logging.getLogger()
        if FFMPEG:
            if fig is None:
                fig = plt.gcf()
            if path is None:
                path = "/tmp/lastMovie.mp4"
            else:
                path = rna.path.resolve(path)
            log.info("Write movie to {path}".format(**locals()))
            with super(Writer, self).saving(fig, path, dpi, **kwargs):
                yield
        else:
            log.warn("FFMPEG Writer not available. Nothing is written!")


class set_zoomable(object):
    """
    Left click a colorbar and release in order to zoom.
    Upper and lower 5% of the colorbar will zoom out.
    """

    def __init__(self, cbar):
        self._cbar = cbar
        self._v_press = None
        self._v_release = None

        artist = self._cbar.mappable
        self._press_connection_id = artist.axes.figure.canvas.mpl_connect(
            "button_press_event", functools.partial(self.on_press)
        )
        self._release_connection_id = artist.axes.figure.canvas.mpl_connect(
            "button_release_event", functools.partial(self.on_release)
        )

    def on_press(self, event):
        if event.inaxes is self._cbar.ax:
            self._v_press = event.ydata

    def on_release(self, event):
        if event.inaxes is self._cbar.ax:
            self._v_release = event.ydata
            self.update_v_min_max()

    def update_v_min_max(self):
        # sort press and release event
        if self._v_press > self._v_release:
            x_up = self._v_press
            x_low = self._v_release
        if self._v_press < self._v_release:
            x_up = self._v_release
            x_low = self._v_press

        # zoom out if in 5% margin
        if x_up > 0.95:
            x_up = 1.5
        if x_low < 0.05:
            x_low = -0.5

        artist = self._cbar.mappable
        vmin, vmax = artist.get_clim()
        v_range = vmax - vmin
        vmax = vmin + x_up * v_range
        vmin = vmin + x_low * v_range
        artist.set_clim(vmin, vmax)
        artist.axes.figure.canvas.draw()


# Backend


class ApiBackendMatplotlib(rna.plotting.ApiBackend):
    """
    # PlotManager "method" not required any more
    """

    EXPLICIT_SAVE_METHODS = {
        ext: "_plt_savefig"
        for ext in plt.gcf().canvas.get_supported_filetypes()
        if ext not in ("pgf")
    }

    def show(self):
        plt.show()

    @staticmethod
    def axes_dim(axes):
        """
        Returns int: axes dimension
        """
        if hasattr(axes, "get_zlim"):
            return 3
        return 2

    @staticmethod
    def gca(dim=None, **kwargs):
        """
        Forwarding to plt.gca but translating the dimension to projection
        correct dimension

        Examples:
            >>> ax_2 = rna.plotting.gca(2)
            >>> assert rna.plotting.axes_dim(ax_2) == 2
            >>> ax_2_implicit = rna.plotting.gca()
            >>> assert rna.plotting.axes_dim(ax_2_implicit) == 2
            >>> assert ax_2 is ax_2_implicit

            >>> ax_3 = rna.plotting.gca(3)
            >>> assert rna.plotting.axes_dim(ax_3) == 3
            >>> ax_3_implicit = rna.plotting.gca()
            >>> assert rna.plotting.axes_dim(ax_3_implicit) == 3
            >>> assert ax_3 is ax_3_implicit

            >>> ax_2 = rna.plotting.gca(2)
            >>> assert rna.plotting.axes_dim(ax_2) == 2

        """
        if dim is not None and "projection" not in kwargs:
            if dim == 2:
                kwargs["projection"] = "rectilinear"
                # the line above would already be good ...
                # but there is a bug in matplotlib which ignores setting the default
                axes = plt.gca(**kwargs)
                if ApiBackendMatplotlib.axes_dim(axes) != dim:
                    return plt.gcf().add_subplot(1, 1, 1, **kwargs)
                else:
                    return axes
            elif dim == 3:
                kwargs["projection"] = "3d"
            else:
                raise NotImplementedError(
                    "Dimension {dim} not implemented".format(**locals())
                )
        axes = plt.gca(**kwargs)
        return axes

    @staticmethod
    def is_axes(obj):
        return isinstance(obj, plt.Axes)

    def set_labels(self, axes, *labels):
        """
        Set x, y and possibly z labels
        """
        axes.set_xlabel(labels[0])
        axes.set_ylabel(labels[1])
        if self.axes_dim(axes) == 3:
            axes.set_zlabel(labels[2])
        elif self.axes_dim(axes) != 2:
            raise ValueError(
                "Axis dimensions other than 2 or 3 is not supported" "for set_labels."
            )

    def set_colorbar(
        self,
        axes,
        artist,
        label=None,
        divide=True,
        position="right",
        size="2%",
        pad=0.05,
        labelpad=None,
        zoom=False,
        **kwargs
    ):
        """
        Note:
            Bug found in matplotlib:
                when calling axes.clear(), the colorbar has to be removed by hand,
                since it will not be removed by clear.
            >> import rna
            >> axes = rna.plotting.gca()
            >> artist = ...
            >> cbar = rna.plotting.set_colorbar(axes, artist)
            >> rna.plotting.save(...)
            >> cbar.remove()  # THIS IS IMPORTANT. Otherwise you will have problems
            # at the next call of savefig.
            >> axes.clear()

        """
        ticks_position = "default"
        label_position = "bottom"
        labelpad = 30 if labelpad is None else labelpad
        if position == "right":
            rotation = 270
        elif position == "left":
            rotation = 90
        elif position == "top":
            rotation = 0
            ticks_position = "top"
            label_position = "top"
            labelpad = 5
        elif position == "bottom":
            rotation = 0
        # colorbar
        if divide and self.axes_dim(axes) == 2:
            divider = pltg1.make_axes_locatable(axes)
            axes = divider.append_axes(position, size=size, pad=pad)
            cbar = plt.colorbar(artist, cax=axes, **kwargs)
        else:
            cbar = plt.colorbar(artist, ax=axes, **kwargs)
        cbar.ax.xaxis.set_ticks_position(ticks_position)
        cbar.ax.xaxis.set_label_position(label_position)
        cbar.ax.tick_params(axis="x", which="major", pad=0)

        # label
        if label is None:
            art_label = artist.get_label()
            if art_label:
                label = art_label
        if label is not None:
            cbar.set_label(label, rotation=rotation, labelpad=labelpad)

        if zoom:
            set_zoomable(cbar)

        return cbar

    @staticmethod
    def clear(axes):
        axes.clear()

    @staticmethod
    def upgrade_style(style, source, dest=None):
        """
        Copy a style file at <origionalFilePath> to the <dest> which is the
        foreseen local matplotlib rc dir by default
        The style will be name <style>.mplstyle
        Args:
            style (str): name of style
            source (str): full path to mplstyle file to use
            dest (str): local directory to copy the file to. Matpotlib has to
                search this directory for mplstyle files!
        Examples:
            >>> import rna
            >>> import os
            >>> rna.plotting.use('matplotlib')
            >>> from rna.plotting.backends.matplotlib import ApiBackendMatplotlib
            >>> ApiBackendMatplotlib.upgrade_style(
            ...     'rna',
            ...     os.path.join(os.path.dirname(rna.plotting.__file__),
            ...                  'mplstyles/rna.mplstyle'))

        """
        if dest is None:
            dest = mpl.get_configdir()
        style_extension = "mplstyle"
        path = rna.path.resolve(os.path.join(dest, style + "." + style_extension))
        source = rna.path.resolve(source)
        rna.path.cp(source, path)

    def set_style(
        self,
        style: str = "rna",
        dest: typing.Optional[str] = None,
        upgrade: bool = False,
    ):
        """
        Set the matplotlib style

        Args:
            style: name of style
            dest: mplstydirectory that stores the mplstyle. if None, use default
                maplotlib destination
            upgrade: if True copy the style to the source destination (dest)
        """
        if dest is None:
            dest = mpl.get_configdir()

        style_extension = "mplstyle"
        path = rna.path.resolve(os.path.join(dest, style + "." + style_extension))
        if style in mpl.style.available and not upgrade:
            plt.style.use(style)
        elif os.path.exists(path) and not upgrade:
            plt.style.use(path)
        else:
            log = logging.getLogger()
            source = rna.path.resolve(
                os.path.join(
                    os.path.join(os.path.dirname(__file__), "mplstyles"),
                    style + "." + style_extension,
                )
            )
            if os.path.exists(source):
                log.warning(
                    "I will copy the style {style} to {dest}.".format(**locals())
                )
                try:
                    self.upgrade_style(style, source, dest)
                    self.set_style(style)
                except Exception:
                    log.error("Could not set style")
            else:
                log.error(
                    "Could not set style {path}. Probably you would want to"
                    "call rna.plotting.upgrade_style(<style>, "
                    "<path to mplstyle file that should be copied>)"
                    "once".format(**locals())
                )

    @contextlib.contextmanager
    def _prepared_figure(self, kwargs):
        """
        Args:
            path
            *args: joined behind path
            **kwargs:
                axes (Axes): specify which axes to save. Higher priority than fig
                fig (Api): specifying which figure should be saved. If none gcf
                width (float): figure width in pts.
                scale (float): Fraction of the width which you wish
                    the figure to occupy
                ratio (float): height to width ratio


        """
        # catch figure from axes or fig
        axes = kwargs.pop("axes", None)
        if axes is None:
            fig_default = plt.gcf()
            axes = self.gca()
            if fig_default is None:
                raise ValueError("fig_default may not be None")
        else:
            fig_default = axes.figure
        fig = kwargs.pop("fig", fig_default)

        # resize the figure to match a given width (in pt).
        width = kwargs.pop("width", None)
        scale = kwargs.pop("scale", 1.0)
        orig_size = fig.get_size_inches()
        orig_ratio = orig_size[1] / orig_size[0]
        ratio = kwargs.pop("ratio", orig_ratio)
        if width is not None:
            fig.set_size_inches(*rna.plotting.figsize(width, scale=scale, ratio=ratio))

        # crop the plot down based on the extents of the artists in the plot
        kwargs["bbox_inches"] = kwargs.pop("bbox_inches", "tight")
        if kwargs["bbox_inches"] == "tight":
            extra_artists = None
            for ax in fig.get_axes():
                first_label = ax.get_legend_handles_labels()[0] or None
                if first_label:
                    if not extra_artists:
                        extra_artists = []
                    if isinstance(first_label, list):
                        extra_artists.extend(first_label)
                    else:
                        extra_artists.append(first_label)
            kwargs["bbox_extra_artists"] = kwargs.pop(
                "bbox_extra_artists", extra_artists
            )

        # set current figure
        plt.figure(fig.number)

        yield

        # reset the figure size
        if width is not None:
            fig.set_size_inches(*orig_size)

    def _plt_savefig(self, path, **kwargs):
        with self._prepared_figure(kwargs):
            plt.savefig(path, **kwargs)

    def _save_tikz(self, path, **kwargs):
        width = kwargs.get("width", None)
        scale = kwargs.get("scale", 1.0)

        with self._prepared_figure(kwargs):
            import matplotlib2tikz as tikz  # pylint: disable=

            if width is None:
                figurewidth = r"\textwidth"
            else:
                figurewidth = "{0}pt".format(scale * width)
            tikz.save(path, figurewidth=figurewidth)

    def _save_pgf(self, path, **kwargs):
        """
        Args:
            **kwargs:
                post_process (bool): switch post-processing of pgfs on and of
                substitutes (list of tuple of str):
                scaling (str): Default is \\igraph@width
                - further are passed to savefig and _pre_save
        """
        post_process = kwargs.pop("post_process", False)
        substitutes = kwargs.pop("substitutes", [(r"\sffamily", "")])
        scaling = kwargs.pop("scaling", r"\igraph@width")

        with self._prepared_figure(kwargs):
            plt.savefig(path, **kwargs)

            # pgf post-processing works but matpltolib pgf backend writes pgftext left
            # adjusted and creates right adjusted text by moving the text in x
            # this makes it impossible to rescale text correctly. If one knew, the
            # original was right adjusted a correction could be applied. This way
            # however, we cann not do anything.
            if post_process and path.endswith("pgf"):
                # use it like \igraph[width=0.5\columnwidth]{path_to_the.pgf}
                with open(path, "r") as f:
                    content = []
                    exp = r"(\\pgftext.+)(\\fontsize{.+)\\selectfont(.+)}%"
                    exp_width = (
                        r"\\pgfpathrectangle{\\pgfpointorigin}"
                        r"{\\pgfqpoint{(.+)}{(.+)}%"
                    )
                    for line in f:
                        res = re.match(exp, line)
                        res_width = re.match(exp_width, line)
                        if res_width:
                            pgf_path_size = res_width.groups()[0]
                            # factor 72.27 (latex definition differs from pdf
                            # definition) for conversion from in to pt
                            pgf_path_size = (
                                r"(" + pgf_path_size.replace("in", "*72.27") + r")"
                            )
                        if res:
                            groups = res.groups()
                            line = (
                                groups[0]
                                + r"\selectfont \scalebox{\fpeval{(%s/(%s))}}{"
                                % (pgf_path_size, scaling)
                                + groups[2]
                                + r"}}%"
                                + "\n"
                            )
                        for x, y in substitutes:
                            line = line.replace(x, y)
                        content.append(line)
                with open(path, "w") as f:
                    f.writelines(content)

    @rna.plotting.plot_signature
    def plot_function(self, axes, fun, **kwargs):
        """
        Args:
            axes (matplotlib.Axis) object
            fun: callable
            **kwargs:
                num: degree of detail (third argument of linspace)
                futher are forwarded to plot

        Returns:
            Artist or list of Artists (imitating the axes.scatter/plot behaviour).
            Better Artist not list of Artists
        """
        import numpy as np

        xmin, xmax = kwargs.pop("xmin", 0), kwargs.pop("xmax", 1)
        num = kwargs.pop("num", 100)
        np.linspace
        vals = np.linspace(xmin, xmax, num)
        args = (vals, map(fun, vals))
        artist = axes.plot(*args, **kwargs)
        return artist

    @rna.plotting.plot_signature
    def plot_errorbar(self, axes, points, errors_up, errors_down=None, **kwargs):
        """
        Args:
            axes (matplotlib.Axis) object

        Returns:
            Artist or list of Artists (imitating the axes.scatter/plot behaviour).
            Better Artist not list of Artists
        """
        po = rna.plotting.PlotManager(kwargs)
        po.setdefault("marker", "_")

        if errors_down is None:
            errors_down = errors_up

        artists = []

        # plot errorbars
        for i in range(len(points)):
            artists.append(
                axes.plot(
                    [points[i, 0] + errors_up[i, 0], points[i, 0] - errors_down[i, 0]],
                    [points[i, 1], points[i, 1]],
                    [points[i, 2], points[i, 2]],
                    **kwargs
                )
            )
            artists.append(
                axes.plot(
                    [points[i, 0], points[i, 0]],
                    [points[i, 1] + errors_up[i, 1], points[i, 1] - errors_down[i, 1]],
                    [points[i, 2], points[i, 2]],
                    **kwargs
                )
            )
            artists.append(
                axes.plot(
                    [points[i, 0], points[i, 0]],
                    [points[i, 1], points[i, 1]],
                    [points[i, 2] + errors_up[i, 2], points[i, 2] - errors_down[i, 2]],
                    **kwargs
                )
            )

        return artists

    @rna.plotting.plot_signature
    def plot_tensor(
        self,
        axes,
        points: np.ndarray,
        field: typing.Optional[np.ndarray] = None,
        **kwargs
    ):
        """
        Args:
            points: base vectors
            field: direction field
            **kwargs: passed to PlotManager
        """
        x_index, y_index, z_index = self.pop_xyz_index(kwargs)

        if field is not None and len(field.shape) == 2 and field.shape[1] == 1:
            # scalar
            field = field.reshape(len(points))

        if field is None or len(field.shape) == 1:
            # scalar
            if self.axes_dim(axes) == 2:
                points = [points[:, x_index], points[:, y_index]]
            else:
                points = [points[:, x_index], points[:, y_index], points[:, z_index]]
            color = kwargs.pop("color", field)
            artists = axes.scatter(*points, c=color, **kwargs)
        elif len(field.shape) == 2:
            # vector
            if points is None:
                points = np.full(field.shape, 0.0)
            artists = []
            for point, vector in zip(points, field):
                if self.axes_dim(axes) == 3:
                    artists.append(
                        axes.quiver(
                            point[x_index],
                            point[y_index],
                            point[z_index],
                            vector[x_index],
                            vector[y_index],
                            vector[z_index],
                            **kwargs
                        )
                    )
                elif self.axes_dim(axes) == 2:
                    artists.append(
                        axes.quiver(
                            point[x_index],
                            point[y_index],
                            vector[x_index],
                            vector[y_index],
                            **kwargs
                        )
                    )
                else:
                    raise NotImplementedError("Dimension != 2|3")
        else:
            raise NotImplementedError("Only Scalars and Vectors implemented")
        return artists

    @rna.plotting.plot_signature
    def plot_mesh(self, axes, vertices, faces, **kwargs):
        """
        Args:
            axes (matplotlib axes)
            x_index (int)
            y_index (int)
            z_index (int)
            edgecolor (color)
            color (color): if given, use this color for faces in 2D
            cmap
            vmin
            vmax
        """
        x_index, y_index, z_index = self.pop_xyz_index(kwargs)
        vertices = np.array(vertices)
        faces = np.array(faces)
        if faces.shape[0] == 0:
            logging.warning("No faces to plot")
            return None
        if max(faces.flat) > vertices.shape[0]:
            raise ValueError("Some faces point to non existing vertices.")

        dim = self.axes_dim(axes)
        if dim == 2:
            full = True
            import tfields

            mesh = tfields.Mesh3D(vertices, faces=faces)
            facecolors = self.retrieve(
                kwargs, "facecolors", "color", default=0, keep=False
            )
            if full:
                # implementation that will sort the triangles by z_index
                centroids = mesh.centroids()
                axis_indices = [0, 1, 2]
                axis_indices.pop(axis_indices.index(x_index))
                axis_indices.pop(axis_indices.index(y_index))
                z_index = axis_indices[0]
                zs = centroids[:, z_index]
                try:
                    # no numbers
                    iter(facecolors)
                    # no strings
                    if isinstance(facecolors, str):
                        raise TypeError()
                except TypeError:
                    zs, faces = tfields.lib.util.multi_sort(zs, faces)
                else:
                    zs, faces, facecolors = tfields.lib.util.multi_sort(
                        zs, faces, facecolors
                    )

                n_faces_initial = len(faces)
            else:
                # cut away "back sides" implementation
                direction_vector = np.array([1.0, 1.0, 1.0])
                direction_vector[x_index] = 0.0
                direction_vector[y_index] = 0.0
                norm_vectors = mesh.triangles().norms()
                dot_product = np.dot(norm_vectors, direction_vector)
                n_faces_initial = len(faces)
                faces = faces[dot_product > 0]

            vertices = [mesh[:, x_index], mesh[:, y_index]]

            """
            sort out color arguments
            """
            facecolors = self.format_colors(
                kwargs, facecolors, fmt="norm", length=n_faces_initial
            )
            if not full:
                facecolors = facecolors[dot_product > 0]

            artist = axes.tripcolor(
                *vertices, triangles=faces, facecolors=facecolors, **kwargs
            )
        elif dim == 3:
            label = kwargs.pop("label", None)
            color = self.retrieve(
                kwargs, "color", "c", "facecolors", default="grey", keep=False
            )
            from rna.plotting.colors import color_fmt

            orig_fmt = color_fmt(color)
            if orig_fmt == "norm":
                orig_color = self.format_colors(
                    kwargs, color, fmt="norm", length=len(faces)
                )

            color = self.format_colors(
                kwargs, color, fmt="rgba", length=len(faces), dtype=float
            )

            nan_mask = np.isnan(color)
            if nan_mask.any():
                logging.warning("nan found in colors. Removing corresponding faces!")
                color = color[~nan_mask]
                faces = faces[~nan_mask]

            edgecolor = kwargs.pop("edgecolor", None)
            alpha = kwargs.pop("alpha", None)
            self.pop_norm_args(kwargs)

            triangles = np.array([vertices[face] for face in faces])
            artist = plt3d.art3d.Poly3DCollection(triangles, **kwargs)
            axes.add_collection3d(artist)

            if edgecolor is not None:
                artist.set_edgecolor(edgecolor)
                artist.set_facecolors(color)
            else:
                artist.set_color(color)
            if orig_fmt == "norm":
                # sets vmin, vmax to min, max of val
                artist.set_array(np.array(orig_color))
            if alpha is not None:
                artist.set_alpha(alpha)

            # for some reason auto-scale does not work
            self.autoscale_3d(axes, array=vertices)

            # legend lables do not work at all as an argument
            if label:
                artist.set_label(label)

            # when plotting the legend edgecolors/facecolors2d are needed
            artist._edgecolors2d = None
            artist._facecolors2d = None

        else:
            raise NotImplementedError("Dimension != 2|3")

        return artist

    @rna.plotting.plot_signature
    def plot_sphere(self, axes, point, radius, **kwargs):
        # Make data
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x = point[0] + radius * np.outer(np.cos(u), np.sin(v))
        y = point[1] + radius * np.outer(np.sin(u), np.sin(v))
        z = point[2] + radius * np.outer(np.ones(np.size(u)), np.cos(v))

        # Plot the surface
        return axes.plot_surface(x, y, z, **kwargs)

    @rna.plotting.plot_signature
    def plot_plane(self, axes, point, normal, **kwargs):
        def rotation_matrix(d):
            sin_angle = np.linalg.norm(d)
            if sin_angle == 0:
                return np.identity(3)
            d /= sin_angle
            eye = np.eye(3)
            ddt = np.outer(d, d)
            skew = np.array(
                [[0, d[2], -d[1]], [-d[2], 0, d[0]], [d[1], -d[0], 0]], dtype=np.float64
            )

            M = ddt + np.sqrt(1 - sin_angle ** 2) * (eye - ddt) + sin_angle * skew
            return M

        def pathpatch_2d_to_3d(pathpatch, z, normal):
            if type(normal) is str:  # Translate strings to normal vectors
                index = "xyz".index(normal)
                normal = np.roll((1.0, 0, 0), index)

            normal /= np.linalg.norm(normal)  # Make sure the vector is normalised
            path = pathpatch.get_path()  # Get the path and associated transform
            trans = pathpatch.get_patch_transform()

            path = trans.transform_path(path)  # Apply the transform

            pathpatch.__class__ = plt3d.art3d.PathPatch3D  # Change the class
            pathpatch._code3d = path.codes  # Copy the codes
            pathpatch._facecolor3d = pathpatch.get_facecolor  # Get the face color

            verts = path.vertices  # Get the vertices in 2D

            d = np.cross(normal, (0, 0, 1))  # Obtain the rotation vector
            M = rotation_matrix(d)  # Get the rotation matrix

            pathpatch._segment3d = np.array(
                [np.dot(M, (x, y, 0)) + (0, 0, z) for x, y in verts]
            )

        def pathpatch_translate(pathpatch, delta):
            pathpatch._segment3d += delta

        kwargs["alpha"] = kwargs.pop("alpha", 0.5)
        patch = mplpatch.Circle((0, 0), **kwargs)
        axes.add_patch(patch)
        pathpatch_2d_to_3d(patch, z=0, normal=normal)
        pathpatch_translate(patch, (point[0], point[1], point[2]))

    @rna.plotting.plot_signature
    def autoscale_3d(self, axes, array=None, xLim=None, yLim=None, zLim=None):
        if array is not None:
            xMin, yMin, zMin = array.min(axis=0)
            xMax, yMax, zMax = array.max(axis=0)
            xLim = (xMin, xMax)
            yLim = (yMin, yMax)
            zLim = (zMin, zMax)
        xLimAxis = axes.get_xlim()
        yLimAxis = axes.get_ylim()
        zLimAxis = axes.get_zlim()

        if not False:
            # not empty axes
            xMin = min(xLimAxis[0], xLim[0])
            yMin = min(yLimAxis[0], yLim[0])
            zMin = min(zLimAxis[0], zLim[0])
            xMax = max(xLimAxis[1], xLim[1])
            yMax = max(yLimAxis[1], yLim[1])
            zMax = max(zLimAxis[1], zLim[1])
        axes.set_xlim([xMin, xMax])
        axes.set_ylim([yMin, yMax])
        axes.set_zlim([zMin, zMax])

    @rna.plotting.plot_signature
    def set_legend(self, axes, *artists, **kwargs):
        """
        Convenience method to set a legend from multiple artists to an axes.
        Args:
            **kwargs
                table (bool): if True, labels containing ',' are mapped to table
                table_title (str): value of the table entry top left - only active
                    if table
        Examples:
            >>> import rna
            >>> import matplotlib.pyplot as plt
            >>> import pylab

            >>> fig = plt.figure()
            >>> ax = fig.add_subplot(111)

            >>> im1 = ax.plot(range(10), pylab.randn(10), "r--",
            ...              label=(r"$i = 1$,$j = 1$"))
            >>> im2 = ax.plot(range(10), pylab.randn(10), "g--",
            ...              label=(r"$i = 1$,$j = 2$"))
            >>> im3 = ax.plot(range(10), pylab.randn(10), "b--",
            ...              label=(r"$i = 1$,$j = 3$"))
            >>> im4 = ax.plot(range(10), pylab.randn(10), "r.",
            ...              label=(r"$i = 2$,$j = 1$"))
            >>> im5 = ax.plot(range(10), pylab.randn(10), "g.",
            ...              label=(r"$i = 2$,$j = 2$"))
            >>> im6 = ax.plot(range(10), pylab.randn(10), "b.",
            ...              label=(r"$i = 2$,$j = 3$"))
            >>> im7 = ax.plot(range(10), pylab.randn(10), "r^",
            ...              label=(r"$i = 3$,$j = 1$"))
            >>> im8 = ax.plot(range(10), pylab.randn(10), "g^",
            ...              label=(r"$i = 3$,$j = 2$"))
            >>> im9 = ax.plot(range(10), pylab.randn(10), "b^",
            ...              label=(r"$i = 3$,$j = 3$"))
            >>> handles = [im1, im2, im3, im4, im5, im6, im7, im8, im9]

            >>> legend = rna.plotting.set_legend(*handles, table=True, axes=ax)

            >>> rna.plotting.show()
        """
        table = kwargs.pop("table", False)
        labels = kwargs.pop("labels", None)
        ncol = kwargs.pop("ncol", None)

        handles = []
        for artist in artists:
            if isinstance(artist, list):
                handles.append(artist[0])
            elif isinstance(artist, tuple):
                tuple_handle = tuple()
                for sub_artist in artist:
                    if isinstance(sub_artist, list):
                        tuple_handle += (sub_artist[0],)
                    else:
                        tuple_handle += (sub_artist,)
                handles.append(tuple_handle)
            else:
                handles.append(artist)

        if labels is None and any([isinstance(h, tuple) for h in handles]):
            labels = []
            for h in handles:
                if isinstance(h, tuple):
                    sub_labels = [sub_h.get_label() for sub_h in h]
                    label = sub_labels[0]
                    for sub_label in sub_labels[1:]:
                        if not sub_label.startswith("_"):
                            label += sub_label
                else:
                    label = h.get_label()
                labels.append(label)

        if table and labels is None and ncol is None:
            table_title = kwargs.pop("table_title", "")
            labels = np.array([h.get_label() for h in handles])
            labels = [label.split(",") for label in labels]
            captions_i = []
            captions_j = []
            for label in labels:
                if label[0] not in captions_i:
                    captions_i.append(label[0])
                if label[1] not in captions_j:
                    captions_j.append(label[1])

            shape = (len(captions_i), len(captions_j))

            # initialize
            shape = np.array(shape)
            handles = np.array(handles)

            # create blank rectangle
            extra = mplpatch.Rectangle(
                (0, 0), 1, 1, fc="w", fill=False, edgecolor="none", linewidth=0
            )

            # Create organized list containing all handles for table.
            # Extra represent empty space
            handles_table = np.full(shape + 1, extra, dtype=object)
            for handle, label in zip(handles, labels):
                i = captions_i.index(label[0])
                j = captions_j.index(label[1])
                if handles_table[i + 1, j + 1] != extra:
                    raise ValueError("Duplicate label {label}".format(**locals()))
                handles_table[i + 1, j + 1] = handle

            # Define the label captions
            labels_table = np.full(shape + 1, "", dtype="S80")
            labels_table[0, 0] = table_title
            labels_table[0, 1:] = captions_j
            labels_table[1:, 0] = captions_i
            labels_table = labels_table.astype(str)

            handles = list(handles_table.flat)
            labels = list(labels_table.flat)
            kwargs["ncol"] = shape[0] + 1
            # negative numbers in handletextpad move to the right
            kwargs["handletextpad"] = kwargs.pop("handletextpad", -1.5)
            kwargs["columnspacing"] = kwargs.pop("columnspacing", 1.5)

        return axes.legend(handles=handles, labels=labels, **kwargs)

    def set_axis_off(self, axes):
        """
        Turn off axis display
        """
        if self.axes_dim(axes) == 2:
            axes.set_axis_off()
        else:
            axes._axis3don = False

    def set_aspect_equal(self, axes):
        """Fix equal aspect bug for 3D plots."""

        if self.axes_dim(axes) == 2:
            axes.set_aspect("equal")
            return

        xlim = axes.get_xlim3d()
        ylim = axes.get_ylim3d()
        zlim = axes.get_zlim3d()

        from numpy import mean

        xmean = mean(xlim)
        ymean = mean(ylim)
        zmean = mean(zlim)

        plot_radius = max(
            [
                abs(lim - mean_)
                for lims, mean_ in ((xlim, xmean), (ylim, ymean), (zlim, zmean))
                for lim in lims
            ]
        )

        axes.set_xlim3d([xmean - plot_radius, xmean + plot_radius])
        axes.set_ylim3d([ymean - plot_radius, ymean + plot_radius])
        axes.set_zlim3d([zmean - plot_radius, zmean + plot_radius])

    def set_limits(self, axes, *limits):
        """
        Args:
            *limits: list of 2-tuples containing min and max limit of axes
        """
        dim = self.axes_dim(axes)
        for coord, limit in zip(["x", "y", "z"][:dim], limits):
            if dim == 2:
                set_method = getattr(axes, "set_" + coord + "lim")
            elif dim == 3:
                set_method = getattr(axes, "set_" + coord + "limd3")

            set_method(*limit)  # limit = xmin, xmax

    @rna.plotting.plot_signature
    def set_formatter(
        self, axes, sub_axes=None, formatter=dates.DateFormatter("%d-%m-%y")
    ):
        """
        Examples:
            >>> from rna.plotting.backends.matplotlib import ApiBackendMatplotlib
            >>> axes = ApiBackendMatplotlib().gca()
            >>> ApiBackendMatplotlib().set_formatter(axes)
        """
        if sub_axes is None:
            sub_axes = axes.xaxis
        sub_axes.set_major_formatter(formatter)


class ScientificFormatter(mpl.ticker.ScalarFormatter):
    """
    Examples:
        >> cbar = rna.plotting.set_colorbar(
        ...     axes, artist,
        ...     label=r"$q_c\;(MW/m^2)$",  # NOQA
        ...     format=rna.plotting.backend.ScientificFormatter(
        ...         None, useMathText=False))

    Args:
        oom (int): order of magnitued on the axes
        **kwargs: forwarded to ScalarFormatter

    """

    def __init__(self, oom=None, **kwargs):
        self._oom = oom
        super().__init__(**kwargs)

    def _set_orderOfMagnitude(self, oom):  # pylint: disable=invalid-name
        exp = int(np.log10(oom))
        if self._oom is not None:
            oom = self._oom
        else:  # Default: -3, 0, 3, 6, ...
            oom = exp - (exp % 3)
        self.orderOfMagnitude = oom



#!/user/bin/env python
"""
w7x option starter
"""
# import sys
# import argparse
# import rna
#
#
# class SomeAction(argparse.Action):
#     def __init__(self, option_strings, dest, nargs=None, **kwargs):
#         if nargs is not None:
#             raise ValueError("nargs not allowed")
#         super().__init__(option_strings, dest, **kwargs)
#
#     def __call__(self, parser, namespace, values, option_string=None):
#         print("Example action invoked by manage in namespace: %r with values %r"
#               " and option string %r" % (namespace, values, option_string))
#         setattr(namespace, self.dest, values)
#
#
# def manage(args):
#     print("Managing!")
#     print(args.x * args.y)
#
#
# def parse_args(args):
#     # create the top-level parser
#     parser = argparse.ArgumentParser(prog='rna app')
#     parser.add_argument('--version', action='version',
#                         version='v' + rna.__version__,
#                         help="Show program's version number and exit")
#     parser = argparse.ArgumentParser(prog='rna app')
#
#     # subparsers
#     subparsers = parser.add_subparsers(help='sub-command help')
#
#     # create the parser for the "test" command
#     example_sub_parser = subparsers.add_parser('manage', help='manage something')
#     example_sub_parser.add_argument('-x', type=int, default=1)
#     example_sub_parser.add_argument('-y', type=float, default=42.)
#     example_sub_parser.set_defaults(func=manage)
#
#     # If no arguments were used, print base-level help with possible commands.
#     if len(args) == 0:
#         parser.print_help(file=sys.stderr)
#         sys.exit(1)
#
#     args = parser.parse_args(args)
#     # let argparse do the job of calling the appropriate function after
#     # argument parsing is complete
#     args.func(args)
#
#
# if __name__ == '__main__':
#     args = parse_args(sys.argv[1:])



"""
Color conversion
"""
from six import string_types
from itertools import cycle
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


FMT_DTYPES = {
    # 'hex': np.dtype([('hex', '<U7')]),
    "rgba": np.dtype([("r", "<f8"), ("g", "<f8"), ("b", "<f8"), ("a", "<f8")]),
    "rgb": np.dtype([("r", "<f8"), ("g", "<f8"), ("b", "<f8")]),
}


def color_fmt(colors):
    """
    Determine the color representation type
    Definition confirms to https://matplotlib.org/api/colors_api.html

    Args:
        colors (various)

    Returns:
        str | None: color representation type
            'str' -> predefined colors e.g. 'r', 'g', 'k', ...
            'gray' -> grayscale value e.g. '0.5'
            'norm' -> arbitrary numbers e.g. [1,2,3.42]
            'rgba' -> tuples of (red, green, blue, transparency)
            'rgb' -> tuples or (red, green, blue)
            None -> fmt unknown

    """
    if isinstance(colors, np.ndarray):
        for fmt, dtype in FMT_DTYPES.items():
            if dtype == colors.dtype:
                return fmt

    if not hasattr(colors, "__iter__") or isinstance(
        colors, (string_types, tuple)
    ):  # NOQA
        # colors is just one element
        colors = [colors]
    color = colors[0]
    norm_types = (int, float, np.float32, np.float64, np.integer)

    if isinstance(color, string_types):
        try:
            c = eval(color)
        except Exception:
            pass
        else:
            if isinstance(c, norm_types):
                return "gray"
        if color.startswith("#"):
            return "hex"
        return "str"
    if isinstance(color, tuple):
        if len(color) == len(FMT_DTYPES["rgba"]):
            return "rgba"
        if len(color) == len(FMT_DTYPES["rgb"]):
            return "rgb"
    if isinstance(color, norm_types):
        return "norm"
    return None


def truncate_colormap(cmap, vmin=0.0, vmax=1.0, n=100):
    cmap = plt.get_cmap(cmap)
    new_cmap = mpl.colors.LinearSegmentedColormap.from_list(
        "trunc({n},{a:.2f},{b:.2f})".format(n=cmap.name, a=vmin, b=vmax),
        cmap(np.linspace(vmin, vmax, n)),
    )
    return new_cmap


def to_colors(
    colors, fmt="rgba", length=None, vmin=None, vmax=None, cmap=None, dtype=None
):
    """
    format colors according to fmt argument
    Args:
        colors (List(tuple|int|float|str)): rgb|rgba|string representation|hex
            interpreted as color
        fmt (str): one of rgba, rgb, hex or norm
        length (int/None): if not None: correct colors lenght
        vmin (float)
        vmax (float)
        cmap (str|mpl.cm.ColorMap)
        dtype (np.dtype): output data type

    Returns:
        colors in fmt

    Examples:
        >>> import rna
        >>> import numpy as np

        >>> rna.plotting.colors.to_colors(1, cmap='gray')
        array((1., 1., 1., 1.),
              dtype=[('r', '<f8'), ('g', '<f8'), ('b', '<f8'), ('a', '<f8')])
        >>> rna.plotting.colors.to_colors(1, fmt='rgb', length=2, cmap='gray')
        array([(1., 1., 1.), (1., 1., 1.)],
              dtype=[('r', '<f8'), ('g', '<f8'), ('b', '<f8')])
        >>> rna.plotting.colors.to_colors([1], cmap='gray', vmin=0)
        array([(1., 1., 1., 1.)],
              dtype=[('r', '<f8'), ('g', '<f8'), ('b', '<f8'), ('a', '<f8')])
        >>> rna.plotting.colors.to_colors('k', cmap='gray', vmin=0, vmax=0)
        array((0., 0., 0., 1.),
              dtype=[('r', '<f8'), ('g', '<f8'), ('b', '<f8'), ('a', '<f8')])
        >>> rna.plotting.colors.to_colors(
        ...     ['xkcd:red'], fmt='rgb', length=3, cmap='gray', vmin=0, vmax=0
        ... )  # doctest: +ELLIPSIS
        array([(0.89803922, 0., 0.), (0.898..., 0., 0.), (0.898..., 0., 0.)],
              dtype=[('r', '<f8'), ('g', '<f8'), ('b', '<f8')])
        >>> rna.plotting.colors.to_colors(1, cmap='gray', length=2, fmt='hex')
        array(['#ffffff', '#ffffff'], dtype='<U7')

        Keeps the shape
        >>> rna.plotting.colors.to_colors(np.array([[[1,1],[1,1]]]),
        ...     cmap='gray', vmin=0)
        array([[[(1., 1., 1., 1.), (1., 1., 1., 1.)],
                [(1., 1., 1., 1.), (1., 1., 1., 1.)]]],
              dtype=[('r', '<f8'), ('g', '<f8'), ('b', '<f8'), ('a', '<f8')])

    """
    # mapping also scalars to lists
    has_iter = True
    if not hasattr(colors, "__iter__") or isinstance(colors, (string_types, tuple)):
        # colors is just one element
        has_iter = False
        colors = [colors]

    # work on a flat color array and reshape afterwards
    colors = np.array(colors)
    shape = list(colors.shape)
    colors = np.array(colors.flat)

    # Assume all colors to be of the same fmt
    orig_fmt = color_fmt(colors)

    # check cmap, vmin, vmax for norm conversions
    if orig_fmt == "norm" or fmt == "norm":
        if orig_fmt == "norm":
            if cmap is None:
                raise TypeError("converting from norm always requires cmap")
            if vmin is None:
                vmin = min(colors) if len(set(colors)) > 1 else min(0, min(colors))
            if vmax is None:
                vmax = max(colors) if len(set(colors)) > 1 else max(1, max(colors))

            # already cut initial norm values
            colors[colors < vmin] = vmin
            colors[colors > vmax] = vmax
        else:
            if vmin is None or vmax is None or cmap is None:
                raise TypeError(
                    "converting to norm always requires cmap, " "vmin and vmax."
                )

    # conversion
    if fmt == orig_fmt:
        # already correct
        pass
    elif fmt == "norm":
        # route everything through rgba_to_norm
        colors_rgba = to_colors(colors, "rgba")
        colors = rgba_to_norm(colors_rgba, cmap, vmin, vmax)
    elif fmt == "rgb":
        colors_rgba = to_colors(colors, "rgba", vmin=vmin, vmax=vmax, cmap=cmap)
        colors = colors_rgba[["r", "g", "b"]]
    elif fmt == "rgba":
        if orig_fmt in ("str", "hex"):
            colors = [mpl.colors.to_rgba(color) for color in colors]
        elif orig_fmt == "rgb":
            colors = [tuple(c) + (1,) for c in colors]
        elif orig_fmt == "norm":
            colors = norm_to_rgba(colors, vmin=vmin, vmax=vmax, cmap=cmap)
        else:
            raise NotImplementedError(
                "Conversion from '{orig_fmt}' to rgba"
                " not implemented.".format(**locals())
            )
    elif fmt == "hex":
        colors_rgba = to_colors(colors, "rgba", vmin=vmin, vmax=vmax, cmap=cmap)
        colors = [mpl.colors.to_hex(tuple(c)) for c in colors_rgba]
    else:
        raise NotImplementedError(
            "Color fmt '{fmt}' not implemented.".format(**locals())
        )

    # set correct length
    if length is not None:
        # just one colors value given
        if len(colors) != length:
            if len(colors) != 1:
                raise ValueError("Can not correct color length")
            colors = [colors[0] for i in range(length)]
            # colors = np.repeat(colors, length, axis=0)  # this does not cast
            # to structured array properly
            shape[0] = length
    elif not has_iter:
        shape = shape[1:]
        colors = colors[0]

    # add fmt info to array dtype
    if fmt in FMT_DTYPES:
        if dtype is None:
            dtype = FMT_DTYPES[fmt]
        else:
            shape.append(len(FMT_DTYPES[fmt]))
    elif hasattr(colors, "dtype"):
        if dtype is None:
            dtype = colors.dtype
        else:
            colors = np.array([c.tolist() for c in colors])

    colors = np.array(colors, dtype=dtype)

    shape = tuple(shape)
    colors = colors.reshape(shape)

    return colors


def norm_to_rgba(scalars, cmap=None, vmin=None, vmax=None):
    """
    retrieve the rgba colors for a list of scalars

    Returns:
        List(tuple)

    Examples:
        >>> import rna
        >>> colors = rna.plotting.colors.norm_to_rgba([0, 1, 1.5, 2, 3, 4],
        ...                                           vmin=1, vmax=2)
        >>> colors = np.array(colors)
        >>> colors = colors[:, :3]  # strip a for test
        >>> assert all(colors[0] == colors[1])
        >>> assert all(colors[1] != colors[2])
        >>> assert all(colors[2] != colors[3])
        >>> assert all(colors[3] == colors[4])
        >>> assert all(colors[4] == colors[5])

    """
    color_map = plt.get_cmap(cmap)
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    colors = color_map([norm(s) for s in scalars])
    colors = [tuple(c) for c in colors]
    return colors


def rgba_to_norm(colors, cmap, vmin, vmax, max_distance=1e-6):
    """
    Inverse 'norm_to_rgba'
    Reconstruct the numeric values (0 - 1) of given
    Args:
        colors (list or rgba tuple)
        cmap (matplotlib colormap)
        vmin (float)
        vmax (float)
        max_distance (float): raise warning if the distance in color space is
            larger than max_distance
    """
    # any to np.ndarray
    colors = np.array(colors)
    # above does not convert struct dtypes. -> struct to float array
    colors = colors.view(np.float64).reshape(colors.shape + (-1,))

    rnge = np.linspace(vmin, vmax, 256)
    norm = mpl.colors.Normalize(vmin, vmax)
    if isinstance(cmap, str):
        cmap = mpl.cm.get_cmap(cmap)
    mapvals = cmap(norm(rnge))[:, :4]  # there are 4 channels: r,g,b,a
    scalars = []
    for color in colors:
        distance = np.sum((mapvals - color) ** 2, axis=1)
        min_distance = np.min(distance)
        if min_distance > max_distance:
            import warnings

            warnings.warn(
                "Distance in color space is large ({min_distance:.6f})".format(  # NOQA
                    **locals()
                )
            )
        scalars.append(rnge[np.argmin(distance)])
    scalars = np.array(scalars)
    return scalars


def colormap(seq):
    """
    Args:
        seq (iterable): a sequence of floats and RGB-tuples.
            The floats should be increasing and in the interval (0,1).
    Returns:
        LinearSegmentedColormap
    """
    seq = [(None,) * 3, 0.0] + list(seq) + [1.0, (None,) * 3]
    cdict = {"red": [], "green": [], "blue": []}
    for i, item in enumerate(seq):
        if isinstance(item, float):
            r1, g1, b1 = seq[i - 1]
            r2, g2, b2 = seq[i + 1]
            cdict["red"].append([item, r1, r2])
            cdict["green"].append([item, g1, g2])
            cdict["blue"].append([item, b1, b2])
    return mpl.colors.LinearSegmentedColormap("CustomMap", cdict)


def color_cycle(cmap=None, n=None):
    """
    Args:
        cmap (matplotlib colormap): e.g. plt.cm.coolwarm
        n (int): number of colors. Only used for cmap argument

    Examples:
        >>> color_cycle('jet', 3)  # doctest: +ELLIPSIS
        <itertools.cycle object at 0x...>
    """
    if cmap:
        if n is None:
            raise AttributeError("color_cycle with cmap requires n")
        colors = to_colors(np.linspace(0, 1, n), cmap=cmap, vmin=0, vmax=1, fmt="hex")
    else:
        colors = list([color["color"] for color in mpl.rcParams["axes.prop_cycle"]])
    return cycle(colors)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Various polymorphisms which can be used as mixins
"""
import pathlib
import typing
import logging
import io
from six import string_types

import rna.path


class Storable:
    """
    Polymorphism with abstract standardized save method and load factory.

    Examples:
        >>> import rna
        >>> class MyStorable(rna.polymorphism.Storable):
        ...     def _save_dummy(self, path):
        ...         print('Saving to', path)
        ...     @classmethod
        ...     def _load_dummy(cls, path):
        ...         print('Loading from', path)
        >>> obj = MyStorable()
        >>> obj.save("/base/path", "to", "obj.dummy", create_parents=False)
        Saving to /base/path/to/obj.dummy
        >>> obj2 = MyStorable.load("/base/path", "to", "obj.dummy")
        Loading from /base/path/to/obj.dummy

    """

    # mapping from extension[str] to method name [str]
    EXPLICIT_SAVE_METHODS: typing.Optional[dict] = None

    def save(
        self,
        path: typing.Union[str, io.IOBase],
        *args: [str],
        create_parents: bool = True,
        extension: typing.Optional[typing.Union[str, typing.List[str]]] = None,
        **kwargs
    ) -> None:
        """
        Saving by redirecting to the correct save method depending on path.
        Implement _save_{extension}(self, path, **kwargs) for saving to your extension of choice

        Args:
            path (str | buffer)
            *args: joined with path
            create_parents: create the parent dir(s) if required
            **kwargs:
                extension (str | List[str]): explicit extension required if path is buffer,
                    if list of extensions save as all extension formats.
                ... remaining:forwarded to extension specific method
        """
        # get the extension
        if isinstance(extension, list):
            for ext in extension:
                print(ext)
                self.save(
                    path, *args, create_parents=create_parents, extension=ext, **kwargs
                )
            return

        if isinstance(path, (string_types, pathlib.Path)):
            path = rna.path.resolve(path, *args)
            if extension is None:
                extension = pathlib.Path(path).suffix.lstrip(".")
            if not path.endswith(extension):
                path += "." + extension
        if not extension:
            raise ValueError("Path requires extension for auto rooting.")

        # get the save method
        if (
            self.EXPLICIT_SAVE_METHODS is not None
            and extension
            in self.EXPLICIT_SAVE_METHODS  # pylint:disable=unsupported-membership-test
        ):
            # pylint:disable=unsubscriptable-object
            method_name = self.EXPLICIT_SAVE_METHODS[extension]
        else:
            method_name = "_save_" + extension
        try:
            save_method = getattr(self, method_name)
        except AttributeError as err:
            raise NotImplementedError(
                "Save method {method_name} for extension: {extension} required.".format(
                    **locals()
                )
            ) from err

        if create_parents:
            rna.path.mkdir(path)

        log = logging.getLogger(__name__)
        log.info("Saving to path '%s' with method %s.", path, save_method)
        save_method(path, **kwargs)

    @classmethod
    def load(cls, path: typing.Union[str, io.IOBase], *args: [str], **kwargs):
        """
        Instantiate an object of this class from file.
        Implement _load_{extension}(cls, path, *args) -> obj for loading from extension of choice.

        Args:
            path (str or buffer)
            *args: joined with path
            **kwargs:
                extension (str): only needed if path is buffer
                ... remaining:forwarded to extension specific method
        """
        if isinstance(path, (string_types, pathlib.Path)):
            # pylint: disable=possibly-unused-variable
            path = rna.path.resolve(path, *args)
            extension = kwargs.pop("extension", pathlib.Path(path).suffix.lstrip("."))
        else:
            extension = kwargs.get("extension")

        if not extension:
            raise ValueError("Path requires extension for auto rooting.")

        try:
            load_method = getattr(cls, "_load_{extension}".format(**locals()))
        except AttributeError as err:
            raise NotImplementedError(
                "Load method '_load_{extension}' for extension: {extension} required.".format(
                    **locals()
                )
            ) from err

        return load_method(path, **kwargs)


class Plottable:  # pylint: disable=too-few-public-methods
    """
    Polymorphism for object which can be plotted
    The polymorphisms lies in the method name *plot' and the suggested kwargs for this method.
    """

    def plot(self, **kwargs):
        """
        Args:
            **kwargs:
                axes: Most fundamental object of the plotting backend e.g. matplotlib.pyplot.Axes
                    Ideally the plot method implements multiple backends which are automatically
                    selected on the basis of the axes object.
                color: color to plot this object. This is to unify the various different signatures
                    of e.g. matplotlib
                cmap (str | colormap object): colorbar map from scalar to color
                vmin: minimum color value for color bar range
                vmax: maximum color value for color bar range
                label(str): label of plotting object
        """
        raise NotImplementedError()



"""
Notes on mapping to matplotlib language:
References:
    * https://pyqtgraph.readthedocs.io/en/latest/plotting.html

Examples:
    $ python -m pyqtgraph.examples

Philosophy:
?       -> QMainWindow

Figure  ->
            LayoutWidget (3d) -> QWidget
            pg.GraphicsLayoutWidget (2D) -> GraphicsWidget -> (GraphicsItem, QGraphicsWidget)
                    -> WINDOW
Axes    -> PlotItem (2D) / GLViewWidget (3D)
Artist  -> GraphicsItem
"""
import typing
import logging
import sys
import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl
import pyqtgraph.exporters
import rna


pg.setConfigOption("background", "w")
pg.setConfigOption("foreground", "w")
pg.setConfigOptions(antialias=True)


global WINDOW, VIEW
WINDOW, VIEW = None, None


def axes_dim(axes):
    if isinstance(axes, pg.PlotItem):
        return 2
    if isinstance(axes, pg.opengl.GLViewWidget):
        return 3
    raise NotImplementedError("Dimension of {axes} not mapped".format(**locals()))


def gca(dim=None):
    global WINDOW, VIEW

    initiate = False

    if VIEW is not None:
        if dim is None:
            return VIEW
        if axes_dim(VIEW) == dim:
            return VIEW
        else:
            initiate = True

    if dim is None:
        dim = 2

    if WINDOW is None:
        pg.mkQApp()
        initiate = True

    if dim == 2:
        VIEW = pg.PlotItem()  # -> GraphicsWidget -> (GraphicsItem, QgraphicsWidget)
        if initiate:
            WINDOW = pg.GraphicsLayoutWidget(title="rna 2D plot")  # show=False
        WINDOW.addItem(VIEW)  # , row, col, rowspan, colspan)

    elif dim == 3:
        VIEW = pg.opengl.GLViewWidget()
        if initiate:
            WINDOW = pg.LayoutWidget()
        WINDOW.addWidget(VIEW)
    else:
        raise NotImplementedError()

    if initiate:
        WINDOW.resize(800, 800)

    return VIEW


def show():
    global WINDOW, VIEW
    WINDOW.show()

    if (sys.flags.interactive != 1) or not hasattr(pg.Qt.QtCore, "PYQT_VERSION"):
        pg.Qt.QtGui.QApplication.instance().exec_()


class GraphicsViewOverlay2D(pg.GraphicsView):
    """
    GraphicsView subclass that uses GLViewWidget as its canvas.
    This allows 2D graphics to be overlaid on a 3D background.
    """

    def __init__(self, axes):
        """
        Args:
            axes (gl.GLViewWidget): view upon which to copy all events
        """
        self.glView = axes
        pg.GraphicsView.__init__(self, background=None)
        self.setStyleSheet("background: transparent")
        self.setViewport(self.glView)

    def mousePos(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mousePos(self, event)
        self.glView.mousePos(event)

    def paintEvent(self, event):
        """
        Distribute paint events to both widgets
        """
        self.glView.paintEvent(event)
        return pg.GraphicsView.paintEvent(self, event)

    def mousePressEvent(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mousePressEvent(self, event)
        self.glView.mousePressEvent(event)

    def mouseMoveEvent(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mouseMoveEvent(self, event)
        self.glView.mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        """
        Distribute mouse events to both widgets
        """
        pg.GraphicsView.mouseReleaseEvent(self, event)
        self.glView.mouseReleaseEvent(event)


class ApiBackendPyqtgraph(rna.plotting.ApiBackend):

    EXPLICIT_SAVE_METHODS = {ext: "_frame_save" for ext in ("png")}

    def show(self):
        show()

    @staticmethod
    def clear(axes):
        global WINDOW, VIEW
        VIEW.setParent(None)
        # WINDOW.removeWidget(VIEW)
        VIEW = None

    @staticmethod
    def gca(dim=None, **kwargs):
        return gca(dim=dim, **kwargs)

    @staticmethod
    def is_axes(obj):
        return isinstance(obj, (pg.PlotItem, pg.opengl.GLViewWidget))

    @staticmethod
    def axes_dim(axes):
        return axes_dim(axes)

    @staticmethod
    def set_colorbar(axes, artist, label=None, position="right", **kwargs):
        # Make a 2D color bar using the same ColorMap
        overlay_view = GraphicsViewOverlay2D(axes)
        kwargs["size"] = kwargs.pop("size", (20, 200))
        kwargs["offset"] = kwargs.pop("offset", (15, -25))
        colorbar_legend = Colorbar(artist, label=label, **kwargs)
        overlay_view.addItem(colorbar_legend)
        return overlay_view

    @staticmethod
    def get_cmap(cm_name):
        """
        Use matplotlib conventions to retrieve a pg.ColorMap

        Examples:
            >>> import rna
            >>> rna.plotting.use('pyqtgraph')
            >>> from rna.plotting.backends.pyqtgraph import ApiBackendPyqtgraph
            >>> cm = ApiBackendPyqtgraph.get_cmap('viridis')
        """
        import matplotlib.pyplot as plt

        plt_map = plt.get_cmap(cm_name)
        colors = plt_map.colors
        colors = [c + [1.0] for c in colors]
        positions = np.linspace(0, 1, len(colors))
        cmap = pg.ColorMap(positions, colors)
        return cmap

    @staticmethod
    def _frame_save(path, **kwargs):
        axes = kwargs.get("axes", None)
        if axes is None:
            axes = gca()
        axes.grabFrameBuffer().save(path, **kwargs)
        # global WINDOW
        # exporter = pg.exporters.ImageExporter(WINDOW.plotItem)
        # exporter.parameters().update(kwargs)
        # exporter.export(path)

    @rna.plotting.plot_signature
    def plot_tensor(
        self,
        axes,
        points: np.ndarray,
        field: typing.Optional[np.ndarray] = None,
        **kwargs
    ):
        """
        Args:
            points: base vectors
            field: direction field
            **kwargs: passed to PlotManager
        """
        raise NotImplementedError(
            "Put a lot of stuff from plot_tensor to ApiBackend and implement"
        )

    @rna.plotting.plot_signature
    def plot_mesh(self, axes, vertices, faces, **kwargs):
        kwargs.setdefault("drawEdges", False)
        kwargs.setdefault("smooth", False)
        kwargs.setdefault("shader", None)
        scalars = kwargs.pop("color", None)
        if scalars is None:
            scalars = [0] * len(faces)
        cmap, vmin, vmax = self.get_norm_args(
            kwargs,
            vmin_default=min(scalars),
            vmax_default=max(scalars),
            cmap_default="viridis",
        )

        colors = self.format_colors(kwargs, scalars, fmt="rgba", length=len(faces))
        colors = np.array([list(c) for c in colors])

        mesh_data = pg.opengl.MeshData(
            vertexes=np.array(vertices), faces=np.array(faces), faceColors=colors
        )
        artist = pg.opengl.GLMeshItem(meshdata=mesh_data, **kwargs)
        artist.colors = colors
        artist.vmin = vmin
        artist.vmax = vmax
        artist.cmap = cmap
        artist.setGLOptions("opaque")

        axes.addItem(artist)
        return artist


class Colorbar(pg.GradientLegend):
    def __init__(self, artist, *args, **kwargs):
        self.label = kwargs.pop("label")
        self.format = kwargs.pop("format", "{0:.2f}")
        self.vmin, self.vmax = artist.vmin, artist.vmax
        ticks = kwargs.pop("ticks", np.linspace(self.vmin, self.vmax, 4))
        super(Colorbar, self).__init__(*args, **kwargs)
        logging.error("Problem with colorbar not yet solved")
        # color_map = get_cmap(artist.cmap)
        # self.setGradient(color_map.getGradient())
        self.set_tick_positions(ticks)

    def set_tick_positions(self, tick_positions):
        if self.vmin > min(tick_positions) or self.vmax < max(tick_positions):
            raise ValueError("tick positions must lie in (vmin, vmax)")

        self.setLabels(
            dict(
                [
                    (
                        self.format.format(v),
                        1 / (self.vmax - self.vmin) * v
                        - self.vmin / (self.vmax - self.vmin),
                    )  # NOQA
                    for v in tick_positions
                ]
            )
        )

    def paint(self, p, opt, widget):
        pg.UIGraphicsItem.paint(self, p, opt, widget)
        rect = self.boundingRect()  # Bounds of visible area in scene coords.
        unit = self.pixelSize()  # Size of one view pixel in scene coords.
        if unit[0] is None:
            return

        # determine max width of all labels
        labelWidth = 0
        labelHeight = 0
        for k in self.labels:
            b = p.boundingRect(
                pg.Qt.QtCore.QRectF(0, 0, 0, 0),
                pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignVCenter,
                str(k),
            )
            labelWidth = max(labelWidth, b.width())
            labelHeight = max(labelHeight, b.height())

        b_tick = p.boundingRect(
            pg.Qt.QtCore.QRectF(0, 0, 0, 0),
            pg.Qt.QtCore.Qt.AlignCenter | pg.Qt.QtCore.Qt.AlignVCenter,
            "-",
        )
        b_label = p.boundingRect(
            pg.Qt.QtCore.QRectF(0, 0, 0, 0),
            pg.Qt.QtCore.Qt.AlignCenter | pg.Qt.QtCore.Qt.AlignVCenter,
            str(self.label),
        )

        labelWidth *= unit[0]
        labelHeight *= unit[1]

        textPadding = 2  # in px

        if self.offset[0] < 0:
            x3 = rect.right() + unit[0] * self.offset[0]
            x2 = x3 - labelWidth - unit[0] * textPadding * 2
            x1 = x2 - unit[0] * self.size[0]
        else:
            x1 = rect.left() + unit[0] * self.offset[0]
            x2 = x1 + unit[0] * self.size[0]
            x3 = x2 + labelWidth + unit[0] * textPadding * 2
        if self.offset[1] < 0:
            y2 = rect.top() - unit[1] * self.offset[1]
            y1 = y2 + unit[1] * self.size[1]
        else:
            y1 = rect.bottom() - unit[1] * self.offset[1]
            y2 = y1 - unit[1] * self.size[1]
        self.b = [x1, x2, x3, y1, y2, labelWidth]

        # Have to scale painter so that text and gradients are correct size.
        p.scale(unit[0], unit[1])

        # Draw color bar
        self.gradient.setStart(0, y1 / unit[1])
        self.gradient.setFinalStop(0, y2 / unit[1])
        p.setBrush(self.gradient)
        rect = pg.Qt.QtCore.QRectF(
            pg.Qt.QtCore.QPointF(x1 / unit[0], y1 / unit[1]),
            pg.Qt.QtCore.QPointF(x2 / unit[0], y2 / unit[1]),
        )
        p.drawRect(rect)

        # draw labels
        p.setPen(pg.Qt.QtGui.QPen(pg.Qt.QtGui.QColor(0, 0, 0)))
        tx = x2 + unit[0] * textPadding
        lh = labelHeight / unit[1]
        for k in self.labels:
            y = y1 + self.labels[k] * (y2 - y1)
            p.drawText(
                pg.Qt.QtCore.QRectF(
                    tx / unit[0] + b_tick.width(), y / unit[1] - lh / 2.0, 1000, lh
                ),
                pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignVCenter,
                str(k),
            )
            p.drawText(
                pg.Qt.QtCore.QRectF(
                    x2,
                    y / unit[1]
                    - b_tick.height() / 2.0
                    - 2,  # the -2 is a hack to move the ticks up  # NOQA
                    b_tick.width(),
                    b_tick.height(),
                ),
                pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignVCenter,
                "-",
            )
        p.save()
        p.translate(x3 + b_label.height(), y1 + 0.5 * (y2 - y1) - b_label.width() / 2)
        p.rotate(90)
        p.drawText(0.0, 0.0, str(self.label))
        p.restore()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Decorator base class providing the same behaviour for both

    @Decorator
    def ..

and

    @Decorator(my_option=True)
    def ..

Additionally it can be applied to methods, function, staticmethods, ... without thinking
"""
import abc
import functools
import inspect


class DecoratorMeta(abc.ABCMeta):
    """
    Meta class for creation of only one instance
    """

    def __call__(cls, *args, **kwargs):
        """
        Allow two signatures:
            1)
                @Decorator
                def fun...

                -> args = (fun,)
                -> kwargs = {}
            2)
                @Decorator(3, option=24)
                def fun...

                -> args = 3
                -> kwargs = {'option': 24}
        """
        func = None
        if args and callable(args[0]) and not args[1:] and not kwargs:
            # Signature 1
            func = args[0]
            args = args[1:]
        instance = cls.__new__(cls, *args, **kwargs)
        instance._func = func  # pylint:disable=attribute-defined-outside-init
        instance._obj = None  # pylint:disable=attribute-defined-outside-init
        instance._wrapped = None  # pylint:disable=attribute-defined-outside-init

        if instance._func is not None:
            # Signature 1
            functools.update_wrapper(instance, func)

        if isinstance(instance, cls):
            instance.__init__(*args, **kwargs)

        return instance


# pylint:disable=access-member-before-definition,attribute-defined-outside-init
class DecoratorBase(metaclass=DecoratorMeta):
    """
    Abstract decorator base class providing the same behaviour for two signatures:
        1)
            @DecoratorBase
            def fun...

            -> args = (fun,)
            -> kwargs = {}
        2)
            @DecoratorBase(3, option=24)
            def fun...

            -> args = 3
            -> kwargs = {'option': 24}

    Additionally it can be applied to methods, function, staticmethods, ... without thinking.
    Additionally functools.wraps is automatically applied to wrapped method. We recommend to
    not use it in addition.
    """

    def __call__(self, *args, **kwargs):
        if self._func is None:
            # Signature 2
            assert len(args) == 1
            assert not kwargs
            self._func = args[0]
            functools.update_wrapper(self, self._func)
            return self.wrapped
        # Signature 1
        return self.wrapped(*args, **kwargs)

    @property
    def wrapped(self):
        """
        Returns:
            wrapped function self._func
        """
        if self._wrapped is None:
            if self._obj:
                # Signature 2: _func is a method
                wrap = self._wrap_method(self._func)
                wrap = functools.partial(wrap, self._obj)
            elif "self" in inspect.Signature.from_callable(self._func).parameters:
                # Signature 1: _func is a method but _obj is None
                wrap = self._wrap_method(self._func)
            else:
                wrap = self._wrap_function(self._func)
            wrap = functools.wraps(self._func)(wrap)
            self._wrapped = wrap
        return self._wrapped

    def __get__(self, obj, obj_type=None):
        # Signature 2: self._func is a method!
        self._obj = obj
        return self

    @abc.abstractmethod
    def _wrap_function(self, func):
        """
        To implement for user

        Returns:
            callable: function
        """

    @abc.abstractmethod
    def _wrap_method(self, func):
        """
        To implement for user

        Returns:
            callable: method
        """


# pylint:disable=too-few-public-methods,invalid-name
class Decorator(DecoratorBase):
    """
    The Decorator class provides easy wrapping with only :py:meth:`_wrap` to be implemented.
    """

    __doc__ = DecoratorBase.__doc__.replace("DecoratorBase", "Decorator") + __doc__

    def _wrap_function(self, func):
        def wrap(*args, **kwargs):
            return self._wrap(None, func, *args, **kwargs)

        return wrap

    def _wrap_method(self, func):
        def wrap(this, *args, **kwargs):
            def prepended_self(*args, **kwargs):
                return func(this, *args, **kwargs)

            return self._wrap(this, prepended_self, *args, **kwargs)

        return wrap

    @abc.abstractmethod
    def _wrap(self, this, func, *args, **kwargs):
        """
        This is a generic wrapper method for both methods and classes. Special about this is that
        it needs to have.

        Args:
            this: obj if func is method then 'this' corresponds to 'self' in the wrapped method.
                if this is None, func is a function or staticmethod
            func: func is the function to be wrapped and will be passed as an explicit kwarg
                NOTE: When calling the function, do not use self as the first attribute even if
                func is a method. This is handled internally.
            *args: Arguments passed to decorated function. Can be passed to func.
            **kwargs: Kwargs passed to decorated function. Can be passed to func.

        Examples:
            >>> from rna.pattern.decorator import Decorator
            >>> # pylint:disable=too-few-public-methods,invalid-name
            >>> class add_args_multiply_kwargs(Decorator):
            ...     def __init__(self, att=None):
            ...         self.att = att
            ...
            ...     def _wrap(
            ...         self, this, func, *args, **kwargs
            ...     ):
            ...         if this is None:  # we are wrapping a function!
            ...             args = args + tuple([sum(args)])
            ...         else:  # we are wrapping a method
            ...             args = args + tuple([this.fast_sum(args)])
            ...         kwargs["res"] = func(*args, **kwargs)
            ...         assert self.att is None or self.att == 42  # we have access to decorator
            ...         return args, kwargs

            >>> class C:
            ...     def fast_sum(self, args):
            ...         return sum(args)
            ...
            ...     @add_args_multiply_kwargs
            ...     def multiply_kwargs(self, *args, **kwargs):
            ...         val = 1
            ...         for v in kwargs.values():
            ...             val *= v
            ...         return val

            >>> @add_args_multiply_kwargs
            ... def multiply_kwargs(*args, **kwargs):
            ...     val = 1
            ...     for v in kwargs.values():
            ...         val *= v
            ...     return val

            >>> C().multiply_kwargs(3, 3, 4, a=2, b=3, c=10)
            ((3, 3, 4, 10), {'a': 2, 'b': 3, 'c': 10, 'res': 60})
            >>> multiply_kwargs(3, 3, 4, a=2, b=3, c=10)
            ((3, 3, 4, 10), {'a': 2, 'b': 3, 'c': 10, 'res': 60})
        """



"""
Strategy design pattern, see https://refactoring.guru/design-patterns/singleton
"""


class SingletonMeta(type):
    """
    Meta class for creation of only one instance
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class Singleton(metaclass=SingletonMeta):  # pylint: disable=too-few-public-methods
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.

    Examples:
        >>> s1 = Singleton()
        >>> s2 = Singleton()
        >>> assert id(s1) == id(s2)
    """



"""Top-level package of rna."""

__author__ = """Daniel Böckenhoff"""
__email__ = "dboe@ipp.mpg.de"
__version__ = "0.7.3"

from . import log  # NOQA
from . import path  # NOQA
from . import process  # NOQA
from . import pattern  # NOQA
from . import polymorphism  # NOQA
from . import plotting

plotting.use("matplotlib")  # TODO: remove



"""
Strategy design pattern, see https://refactoring.guru/design-patterns/strategy

Examples:
    >>> from abc import abstractmethod
    >>> from rna.pattern.strategy import Strategy, Context

    Further user abstraction on top
    >>> class Backend(Strategy):
    ...     @abstractmethod
    ...     def all_backends_should_have_this(self):
    ...         pass

    >>> class Node(Context):
    ...     @property
    ...     def backend(self):
    ...         return self.strategy
    ...
    ...     @backend.setter
    ...     def backend(self, backend):
    ...         self.strategy = backend

    High level description of the still abstact Strategy but the final Context class
    Setting STRATEGY_TYPE defines the Abstract Strategy class every strategy should derive of.
    >>> class MyBackend(Backend):
    ...     @abstractmethod
    ...     def required(self):
    ...         pass
    >>> class MyNode(Node):
    ...     STRATEGY_TYPE = MyBackend
    ...     def compute(self):
    ...         return self.backend.required()

    Finally an implementation of two strategy workers
    >>> class MyBackendVersion1(MyBackend):
    ...     def all_backends_should_have_this(self):
    ...         pass
    ...     def required(self):
    ...         return 42
    >>> class MyBackendVersion2(MyBackend):
    ...     def all_backends_should_have_this(self):
    ...         pass
    ...     def required(self):
    ...         return 21

    >>> a = MyNode(strategy=MyBackendVersion1)
    >>> a.compute()
    42
    >>> a.backend = MyBackendVersion2
    >>> a.compute()
    21

    The backend needs to be a subclass of MyNode.STRATEGY_TYPE
    >>> a.strategy = Backend  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    TypeError: Strategy must be of type <class '...MyBackend'>, not <class '....Backend'>

    Reset the strategy by setting it to None
    >>> a.backend = None

    Context.strategy automatically generates a strategy from STRATEGY_DEFAULT if set.
    >>> MyNode.STRATEGY_DEFAULT = MyBackendVersion1
    >>> a.compute()
    42

    The Context class exposes the abstract strategy methods.
    >>> a.required()
    42
"""
from abc import ABC
import typing
import importlib
import inspect


class Strategy(ABC):
    """
    The Strategy interface declares operations common to all supported versions of some algorithm.

    The Context uses this interface to call the algorithm defined by Concrete Strategies.

    Attributes:
        STRATEGY_MODULE_BASE: If not None given, this is used as a base namespace for the
            from_module method.
        STRATEGY_MODULE_NAME: last part of namespace for the from_module_method.
    """

    STRATEGY_MODULE_BASE: typing.Optional[str] = None  # e.g. "w7x.simulation.backends"
    STRATEGY_MODULE_NAME: typing.Optional[typing.Union[str, int]] = None  # e.g. "flt"

    @classmethod
    def inheritors(cls) -> set:
        """
        Returns:
            flat subclasses and subsub...classes
        """
        subclasses = set()
        work = [cls]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return subclasses

    @classmethod
    def from_module(
        cls, strategy: str, *args, auto_import=True, **kwargs
    ) -> "Strategy":
        """
        Examples:
            >>> from rna.pattern.strategy import Strategy
            >>> backend = Strategy.from_module('rna.pattern.backend')
            >>> from rna.pattern.backend import Backend
            >>> type(backend) is Backend
            True
        """
        # importing the backend should add a subclass of self
        module = ""
        if cls.STRATEGY_MODULE_BASE is not None:
            module = cls.STRATEGY_MODULE_BASE + "."

        module += strategy

        if cls.STRATEGY_MODULE_NAME is not None:
            if isinstance(cls.STRATEGY_MODULE_NAME, int):
                module += "." + cls.__module__.split(".")[cls.STRATEGY_MODULE_NAME]
            elif isinstance(cls.STRATEGY_MODULE_NAME, slice):
                module += "." + ".".join(
                    cls.__module__.split(".")[cls.STRATEGY_MODULE_NAME]
                )
            elif isinstance(cls.STRATEGY_MODULE_NAME, str):
                module += "." + cls.STRATEGY_MODULE_NAME

        for tpe in cls.inheritors():
            if module in tpe.__module__:
                return tpe(*args, **kwargs)

        if auto_import:
            importlib.import_module(module)
            return cls.from_module(strategy, auto_import=False, *args, **kwargs)
        raise ValueError("No subclass of {cls} in module {module}".format(**locals()))


# pylint: disable=too-few-public-methods
class Context:
    """
    The Context defines the interface of interest to clients.

    Args:
        *args: forwarded to strategy (on strategy initialization time - see :meth:`strategy`)
        **kwargs: forwarded to strategy (on strategy initialization time - see :meth:`strategy`)
    """

    STRATEGY_TYPE: Strategy = None  # base class of all allowed strategies
    STRATEGY_DEFAULT: typing.Union[typing.Type[Strategy], str] = None

    def __init__(
        self,
        *args,
        strategy: typing.Optional[
            typing.Union[Strategy, typing.Type[Strategy], str]
        ] = None,
        **kwargs,
    ) -> None:
        """
        Usually, the Context accepts a strategy through the constructor, but
        also provides a setter to change it at runtime.
        """
        self._strategy_init_values = (args, kwargs)
        self.strategy = strategy

    @property
    def strategy(self) -> Strategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.

        Actually instantiate self._strategy with self._strategy_init_values if self._strategy is
        a type instance or string.

        Returns:
            self._strategy: the newly instantiated strategy attribute.
        """
        args, kwargs = self._strategy_init_values
        if self._strategy is None:
            if self.STRATEGY_DEFAULT is None:
                return None
            # pylint:disable=not-callable
            if isinstance(self.STRATEGY_DEFAULT, str):
                self._strategy = self.STRATEGY_TYPE.from_module(
                    self.STRATEGY_DEFAULT, *args, **kwargs
                )
            else:
                self._strategy = self.STRATEGY_DEFAULT(*args, **kwargs)
        elif inspect.isclass(self._strategy) and issubclass(self._strategy, Strategy):
            self._strategy = self._strategy(*args, **kwargs)
        elif isinstance(self._strategy, str):
            self._strategy = self.STRATEGY_TYPE.from_module(
                self._strategy, *args, **kwargs
            )
        return self._strategy

    @strategy.setter
    def strategy(
        self,
        strategy: typing.Union[Strategy, typing.Type[Strategy], str, typing.Type[None]],
    ) -> None:
        """
        Specialty of this setter is that it does also accept uninitialized classes.
        The getter will take care of instantiation. The idea is to not produce any overhead
        until the strategy is actually required.
        """
        if inspect.isclass(strategy) and not issubclass(strategy, self.STRATEGY_TYPE):
            raise TypeError(
                "Strategy must be of type {self.STRATEGY_TYPE}, not {strategy}".format(
                    **locals()
                )
            )
        self._strategy = strategy

    def __getattr__(self, attr: str):
        """
        Forwarding to strategy namespace defined by abstractmethod.
        """
        if hasattr(self.STRATEGY_TYPE, attr):
            value = getattr(self.STRATEGY_TYPE, attr)
            if getattr(value, "__isabstractmethod__", False):
                # Parent strategy defines this attribute as exposed!
                if self.strategy is None:
                    raise AttributeError(
                        "Request to forward '{self.__class__.__name__}' object attribute {attr}"
                        " to un-set strategy.".format(**locals())
                    )
                return getattr(self.strategy, attr)
        raise AttributeError(
            "'{self.__class__.__name__}' object has no attribute '{attr}'".format(
                **locals()
            )
        )



"""
This module is built to be able to use multiple plotting backends in the same
way.
No matter, which backend you use, plotting should work with the steps shown in
the following
Examples:
    >>> import rna
    >>> import tfields

    Build a cube
    >>> cube = tfields.Mesh3D.grid(
    ...         (-1, 1, 2),
    ...         (-1, 1, 2),
    ...         (-5, -3, 2))
    >>> cube.maps[3].fields = [tfields.Tensors(list(range(len(cube.maps[3]))))]
    >>> map_index = 0
    >>> scalars = cube.maps[3].fields[map_index]

    Tell rna to use the pyqtgraph backend. Default is matplotlib
    >>> rna.plotting.use('pyqtgraph')

    Switching back to default for this test. The code below stays the same if you keep pyqtgraph.
    >>> rna.plotting.use('matplotlib')

    Get an axes element
    >>> axes = rna.plotting.gca(3)

    Plot the cube
    >>> artist = rna.plotting.plot_mesh(axes, cube, cube.maps[3], color=scalars,
    ...     cmap='plasma')

    # Plot a colorbar
    >>> cbar = rna.plotting.set_colorbar(axes, artist)  # TODO: Broken in pyqtgraph atm

    Show the app and start the loop. Commented for doctests
    >>> rna.plotting.show()

"""
import typing
import numpy as np
from . import base
from . import colors
from .core import (
    Api,
    ApiBackend,
    plot_signature,
    Figure,
    PlotKwargsFunctions,
)


# TODO: remove EXPOSED_BACKEND_METHOD in favor of check for ApiBackend member
EXPOSED_BACKEND_METHODS = [
    "axes_dim",
    "gca",
    "is_axes",
    "show",
    "clear",
    "plot_mesh",
    "plot_array",
    "plot_tensor",
    "save",
    "set_colorbar",
    "set_style",
    "set_labels",
    "set_legend",
    "set_aspect_equal",
    "set_limits",
]


def use(backend: typing.Union[str, ApiBackend]):
    """
    Activate a backend of you choice. Non specific methods will be exposed to
    the namespace of rna.plotting . Further methods are exposed to
    rna.plotting.backend
    Args:
        newbackend (str):
            The name of the backend to use.
    Examples:
        >>> import rna
        >>> rna.plotting.use('matplotlib')
        >>> mpl_show = rna.plotting.show
        >>> rna.plotting.use('pyqtgraph')
        >>> pqg_show = rna.plotting.show
        >>> assert callable(mpl_show)
        >>> assert callable(pqg_show)
        >>> assert mpl_show is not pqg_show
        >>> rna.plotting.use('matplotlib')
    """
    Api.STRATEGY_DEFAULT = backend


def __getattr__(name: str) -> typing.Any:
    """
    Forward to generic Api methods

    Note:
        Possibly only on 3.7 +
        Maybe look closer to
        https://stackoverflow.com/questions/2447353/getattr-on-a-module
    """
    if name in EXPOSED_BACKEND_METHODS:
        backend = Api().backend
        try:
            return getattr(backend, name)
        except AttributeError as err:
            raise NotImplementedError(
                "Backend {backend} requires implementation of {name}".format(**locals())
            ) from err
    elif not name.startswith("_") and name in dir(PlotKwargsFunctions):
        return getattr(PlotKwargsFunctions, name)
    else:
        raise AttributeError


def figsize(width_pt, scale=1.0, ratio=None):
    """
    Args:
        width_pt (float): figure width width in pt. Can be scaled by
            Get textwidth from latex using '\the\textwidth'
        scale (float)
        ratio (float | None):
            float: ratio of figure height / figure width
            None: golden ratio
    """
    inches_per_pt = 1.0 / 72.27  # Convert pt to inch
    if ratio is None:
        # golden ratio
        ratio = (np.sqrt(5.0) - 1.0) / 2.0
    fig_width = width_pt * inches_per_pt * scale  # fig_width in inches
    fig_height = fig_width * ratio  # height in inches
    fig_size = (fig_width, fig_height)
    return fig_size


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Abstraction of plotting with arbitrary plotting apis (backends)
"""
import typing
import warnings
import abc
import pickle
import rna.pattern.backend
import rna
import rna.pattern.decorator


def _get_axes_from_input(backend, *data, **kwargs):
    axes = kwargs.pop("axes", None)
    dim = kwargs.pop("dim", None)

    if data:
        if backend.is_axes(data[0]):
            if axes is not None and axes is not data[0]:
                raise ValueError("Conflicting axes passed as arg and kwarg")
            axes = data[0]
            data = data[1:]

    if axes and dim:
        if dim != backend.axes_dim(axes):
            raise ValueError(
                "Axes dimension and dim argument in conflict:"
                "{backend.dim} vs {dim}".format(**locals())
            )
    if axes is None:
        axes = backend.gca(dim=dim)
    return axes, data, kwargs


# pylint:disable=invalid-name,too-few-public-methods
class plot_signature(rna.pattern.decorator.Decorator):
    """
    This is a decorator which ensures that you pass an axes instance either
        as first argument
    or
        as keyword arugment axes=<axes instance>
    and/or
        you specify the dimension of the axes object ot be generated dim=<dimension>
    This decorator is applicable to both functions and methods.

    These wraped function requires an axes object as first argument but may be used as if it would
    not exist.

    Notes:
        * Consider https://stackoverflow.com/a/60832711/6233888 to add a backend kwarg

    Examples:
        >>> import rna
        >>> def my_plot_function(axes):
        ...     return axes
        >>> my_decorated_plot_function = rna.plotting.plot_signature(my_plot_function)

        >>> axes = rna.plotting.gca(2)

        Use it without the axes and you have an axes at hand
        >>> assert my_decorated_plot_function() == axes

        You can also pass an axes explicit, both as first argument and as named kwarg
        >>> assert my_decorated_plot_function(rna.plotting.gca()) == axes
        >>> assert my_decorated_plot_function(axes=rna.plotting.gca()) == axes

        Or you give the dimension of the axes
        >>> type(my_decorated_plot_function(dim=3))
        <class 'matplotlib.axes._subplots.Axes3DSubplot'>


    """

    def _wrap(self, this, func, *args, **kwargs):
        if this is None:
            this = rna.plotting
        axes, data, kwargs = _get_axes_from_input(this, *args, **kwargs)
        return func(axes, *data, **kwargs)


class PlotKwargsFunctions:
    """
    Collection of staticmethod functions acting on kwargs passed to a plotting method.
    This provides a standardized signature of plotting functions.
    """

    @staticmethod
    def set(kwargs, attr, value):
        """
        Set attribute with value to kwargs
        """
        kwargs.kwargs[attr] = value

    @staticmethod
    def retrieve(
        kwargs: dict,
        *attrs: str,
        default: typing.Any = None,
        keep: typing.Union[bool, typing.List[str]] = True,
        pop: typing.Optional[typing.List[str]] = None,
        deprecated: typing.Optional[typing.List[str]] = None,
    ):
        """
        Try to find a value in kwargs. Loop through attrs keys and return the first valid
        occurance.
        This can be used e.g. to allow old style arguments where another attribute name should be
        established.

        Args:
            kwargs:
            *attrs: attributes to use as keys for lookup
            default: Default value to return if no value was found
            keep: If False pop the keys from kwargs. If True, keep all.
                If list of attrs -> keep these attributes, remove all others.
            pop: List of attrs to pop. Disjunct with keep so keep is set to False automatically.
            depcreated:

        Examples:
            >>> import rna

            In the simplest case it behaves like get with None as default
            >>> kwargs = dict(a=3)
            >>> rna.plotting.retrieve(kwargs, 'b') is None
            True
            >>> kwargs
            {'a': 3}

            In the second to simplest case it behaves like pop with None as default
            >>> rna.plotting.retrieve(kwargs, 'a', keep=False)
            3
            >>> kwargs
            {}

            Multiple lookups are allowed in a chain
            >>> kwargs = {}
            >>> rna.plotting.retrieve(kwargs, 'edgecolor', 'c', 'color', default='k')
            'k'

            >>> kwargs = dict(color='b', edgecolor='r')
            >>> rna.plotting.retrieve(kwargs, 'color', 'c')
            'b'
            >>> rna.plotting.retrieve(kwargs, 'c', 'color', deprecated=['c'])
            'b'
            >>> rna.plotting.retrieve(kwargs, 'edgecolor', 'c', 'color')
            'r'
            >>> assert 'color' in kwargs
            >>> assert 'edgecolor' in kwargs

            Pop values if no longer required in kwargs:
            >>> rna.plotting.retrieve(kwargs, 'edgecolor', 'c', 'color', keep=False)
            'r'
            >>> assert len(kwargs) == 0

            Pop values except those in the list keep if no longer required in kwargs:
            >>> kwargs = dict(color='b', edgecolor='r', facecolor='g')
            >>> rna.plotting.retrieve(kwargs, 'edgecolor', 'c', 'color', pop=['edgecolor'])
            'r'
            >>> kwargs
            {'color': 'b', 'facecolor': 'g'}
            >>> assert 'color' in kwargs
            >>> assert len(kwargs) == 2

            >>> rna.plotting.retrieve(kwargs, 'facecolor', 'color', keep=False)
            'g'
            >>> assert len(kwargs) == 0

        """
        if pop and (isinstance(keep, list) or not keep):
            raise ValueError("pop and keep arguments are exclusory.")

        found = False
        for attr in attrs:
            if attr not in kwargs:
                continue

            # attr present. pop or get?
            if pop and attr in pop:
                tmp = kwargs.pop(attr)
            elif not keep or (isinstance(keep, list) and attr not in keep):
                tmp = kwargs.pop(attr)
            else:
                tmp = kwargs.get(attr)

            if not found:
                # first occurance
                value = tmp

            found = True

            if not (deprecated or pop or not keep or isinstance(keep, list)):
                # no need to continue
                break

            if deprecated and attr in deprecated:
                warnings.warn(
                    "Attribute {attrs[0]} is deprecated.".format(**locals()),
                    DeprecationWarning,
                )

        if found:
            return value

        return default

    @staticmethod
    def get_norm_args(kwargs, vmin_default=0, vmax_default=1, cmap_default=None):
        """
        Examples:
            >>> import rna
            >>> from rna.plotting.backends.matplotlib import ApiBackendMatplotlib

            >>> rna.plotting.set_style()
            >>> ApiBackendMatplotlib.get_norm_args(dict(vmin=2, vmax=42))
            ('viridis', 2, 42)
        """
        if cmap_default is None:
            import matplotlib.pyplot as plt  # pylint: disable = import-outside-toplevel

            cmap_default = plt.rcParams["image.cmap"]
        cmap = kwargs.get("cmap", cmap_default)
        vmin = kwargs.get("vmin", vmin_default)
        vmax = kwargs.get("vmax", vmax_default)
        if vmin is None:
            vmin = vmin_default
        if vmax is None:
            vmax = vmax_default
        return cmap, vmin, vmax

    @staticmethod
    def pop_norm_args(kwargs, **defaults):
        """
        Pop vmin, vmax and cmap from plot_kwargs
        Args:
            **defaults:
                see get_norm_args method
        """
        cmap, vmin, vmax = ApiBackend.get_norm_args(kwargs, **defaults)
        kwargs.pop("cmap", None)
        kwargs.pop("vmin", None)
        kwargs.pop("vmax", None)
        return cmap, vmin, vmax

    @staticmethod
    def pop_xyz_index(kwargs: dict):
        """
        Args:
            kwargs: contains optional integer keys x_index, y_index and z_index
                Thes indices refer to the index of some data object which holds data of x, y and
                z axis respectively
        """
        x_index = kwargs.pop("x_index", 0)
        y_index = kwargs.pop("y_index", 1)
        z_index = kwargs.pop("z_index", None)
        if z_index is None:
            z_index = {1, 2, 3}
            z_index.difference_update()
            z_index = z_index.pop()
        return x_index, y_index, z_index

    @staticmethod
    def format_colors(kwargs, colors, fmt="rgba", length=None, dtype=None):
        """
        format colors according to fmt argument
        Args:
            colors (list/one value of rgba tuples/int/float/str): This argument
                will be interpreted as color
            fmt (str): rgba | hex | norm
            length (int/None): if not None: correct colors lenght
            dtype (np.dtype): output data type

        Returns:
            colors in fmt
        """
        cmap, vmin, vmax = ApiBackend.get_norm_args(
            kwargs, cmap_default=None, vmin_default=None, vmax_default=None
        )
        from rna.plotting.colors import to_colors

        return to_colors(
            colors, fmt, length=length, vmin=vmin, vmax=vmax, cmap=cmap, dtype=dtype
        )


class ApiBackend(
    rna.pattern.backend.Backend,
    rna.polymorphism.Plottable,
    rna.polymorphism.Storable,
    PlotKwargsFunctions,
):
    """
    Base class for a backend implementation for a Api subclass

    The plotting api backend is thought of as unified interface for plot arguments taking care of
        * interdependent arguments (axes/dimension, cmap/vmin/vmax, x_index,...)
        * color formating (format_colors)

    processing kwargs for plotting functions and providing easy
    access to axes, dimension and plotting method as well as indices
    for array choice (x..., y..., z_index)

    Examples:
        >>> import rna
        >>> from rna.plotting.backends.matplotlib import ApiBackendMatplotlib

        Dimension arguemnt (dim)
        >>> api = rna.plotting.Api()
        >>> api.backend = 'matplotlib'  # which is default anyway

        As a functionality of the Backend/Strategy pattern the backend is instantiated at
        access time
        >>> backend = api.backend
        >>> assert not isinstance(backend, str)
        >>> assert isinstance(backend, ApiBackendMatplotlib)

        Default dimension of the axis is 2
        >>> backend.axes_dim(backend.gca())
        2

        You can select the correct axes object with the dimension key
        >>> axes = ApiBackendMatplotlib.gca(dim=3)
        >>> backend.axes_dim(axes)
        3

        Switching the backend keeps the top level behaviour the same
        >>> api = rna.plotting.Api()
        >>> api.backend = 'pyqtgraph'
        >>> api.backend.axes_dim(api.backend.gca(dim=2))
        2

    """

    STRATEGY_MODULE_BASE = "rna.plotting.backends"

    @plot_signature
    def plot_array(self, *args, **kwargs):
        warnings.warn("Use plot_tensor instead.", DeprecationWarning)
        return self.plot_tensor(*args, **kwargs)  # pylint: disable=no-member

    @staticmethod
    @abc.abstractmethod
    def axes_dim(axes):
        """
        Retrieve the current axes object which resembles the canvas you plot on to

        Args:
            dim: dimension of the axes object
        """

    @staticmethod
    @abc.abstractmethod
    def gca(dim: typing.Optional[int] = None, **kwargs):
        """
        Retrieve the current axes object which resembles the canvas you plot on to

        Args:
            dim: dimension of the axes object
        """

    @staticmethod
    @abc.abstractmethod
    def is_axes(obj: typing.Any) -> bool:
        """
        Check if the given object is an axes instance of this backend
        """

    @abc.abstractmethod
    def show(self):
        """
        Render the current api and promt it to display.
        """


class Api(
    rna.pattern.backend.Frontend, rna.polymorphism.Plottable, rna.polymorphism.Storable
):
    """
    Base class of a generic api.

    Examples:
        >>> import numpy as np
        >>> from rna.plotting import Api
        >>> api = Api()
        >>> api.backend  # doctest +ELLIPSIS
        <rna.plotting.backends.matplotlib.ApiBackendMatplotlib object at ...>

        To use plot, implement _plot
        >>> class Demo(Api):
        ...     def _plot(self, **kwargs):
        ...         self.set_style()
        ...         labels = kwargs.pop("labels", ["x (m)", "y (m)", "z (m)"])
        ...         ax = self.gca()
        ...         artist = self.plot_tensor(ax, self.data, **kwargs)
        ...         self.set_labels(ax, *labels)
        ...         return artist
        >>> demo = Demo()

        >>> artists = demo.plot(np.array([[1,1], [2,2]]), color='r')
        >>> _ = demo.set_legend(artists)

        # saving with a data format stores the data and allows recovery
        >>> path = "./tmp.pkl"
        >>> demo.save(path)
        >>> demo_restored = demo.load(path)

        >>> artists = demo_restored.plot()

        # rendering the api
        >>> demo_restored.show()

        # saving with a api format stores the Api
        >>> path_demo = "./tmp.png"
        >>> demo.save(path_demo)

        The backend can be easily exchanged:
        >>> demo.backend="plotly"
    """

    STRATEGY_TYPE = ApiBackend
    STRATEGY_DEFAULT = "matplotlib"

    def __init__(self):
        # Make sure you pass nothing to backend
        super().__init__()
        self.data = None

    def plot(self, data=None, **kwargs):
        """
        Args:
            data: data object. If you loaded this Api from data file, this is not required
                data is set to self.data which is then available in _plot
            **kwargs: passed to _plot
        """
        if data is None and self.data is None:
            raise ValueError("No data given (first argument)")
        elif data is not None:
            self.data = data
        return self._plot(**kwargs)

    def _plot(self, **kwargs):
        raise NotImplementedError(
            "{self} requires implementation of _plot(**kwargs)".format(**locals())
        )

    def save(self, *args, **kwargs):
        """
        Data is saved from Api class directly.
        Plots are specific to the backend so _save_<ext> must be implemented there. Also they have
        no load method.
        Data has priority if same format is present in Api and backend. Use Api.backend.save
        if you want to save backend also/only in this format.
        """
        try:
            # data saving
            if self.data is None:
                raise ValueError(
                    "{self}.data is None! You forgot to set the data attribute in plotting."
                )
            super().save(*args, **kwargs)  # pylint: disable=no-member
        except NotImplementedError:
            # figure saving
            self.backend.save(*args, **kwargs)

    def _save_pkl(self, path, **kwargs):
        kwargs.setdefault("protocol", pickle.HIGHEST_PROTOCOL)
        with open(path, "wb") as handle:
            pickle.dump(self.data, handle, **kwargs)

    @classmethod
    def _load_pkl(cls, path, **kwargs):
        with open(path, "rb") as handle:
            data = pickle.load(handle, **kwargs)
        obj = cls()
        obj.data = data
        return obj

    def __getattr__(self, name: str) -> typing.Any:
        """
        Forward to generic Api methods either defined by EXPOSED_BACKEND_METHODS or part of
        ApiBackend.

        Note:
            Possibly only on 3.7 +
            Maybe look closer to
            https://stackoverflow.com/questions/2447353/getattr-on-a-module
        """
        if name in rna.plotting.EXPOSED_BACKEND_METHODS:
            backend = (
                self.backend
            )  # CAREFUL: could break the clever logic of the strategy pattern
            try:
                return getattr(backend, name)
            except AttributeError as err:
                raise NotImplementedError(
                    "Backend {backend} requires implementation of {name}".format(
                        **locals()
                    )
                ) from err
        if hasattr(ApiBackend, name):
            return getattr(self.backend, name)
        return self.__getattribute__(name)  # raise AttributeError


class Figure(Api):
    """
    The Figure backend should be derived from
    """

    pass



"""
Alias for strategy pattern. Aliasing makes the use case more understandable in some cases.
"""
from rna.pattern import strategy


class Backend(strategy.Strategy):
    """
    Actual Worker
    """


class Frontend(strategy.Context):
    """
    Client side, deligating work to Backends
    """

    @property
    def backend(self):
        """
        Alias for strategy
        """
        return self.strategy

    @backend.setter
    def backend(self, backend):
        self.strategy = backend



"""
loggin utilities
"""
import typing
from contextlib import contextmanager
import logging
import time


def progressbar(
    iterable: typing.Iterable, logger: typing.Optional[logging.Logger] = None
):
    """
    Args:
        iterable
        logger: optional

    Examples:
        >>> import logging
        >>> import rna
        >>> import sys
        >>> sys.modules['tqdm'] = None
        >>> log = logging.getLogger(__name__)

        >>> a = range(3)
        >>> for value in rna.log.progressbar(a, logger=log):
        ...     _ = value * 3

    """
    if logger is None:
        logger = logging
    try:
        # tqdm not required for the module to work
        from tqdm import tqdm as progressor  # pylint: disable=import-outside-toplevel

        tqdm_exists = True
    except ImportError:

        def progressor(iterable):
            """
            dummy function. Does nothing
            """
            return iterable

        tqdm_exists = False
    try:
        n_total = len(iterable)
    except TypeError:
        n_total = None

    for i in progressor(iterable):
        if not tqdm_exists:
            if n_total is None:
                logger.info("Progress: item {i}".format(**locals()))
            else:
                logger.info("Progress: {i} / {n_total}".format(**locals()))
        yield i


@contextmanager
def timeit(
    process_name: str = "No Description",
    logger: typing.Optional[logging.Logger] = None,
    precision=1,
) -> None:
    """
    Context manager for autmated timeing

    Args:
        process_name (str): message to customize the log message
        logger
        precision (int): show until 10^-<precision> digits

    Examples:
        >>> import rna
        >>> with rna.log.timeit("Logger Name"):
        ...     # Some very long calculation
        ...     pass
    """
    if logger is None:
        logger = logging
    start_time = time.time()
    message = "Timing Process: {0} ...".format(process_name)
    logger.log(logging.INFO, message)

    yield

    logger.log(
        logging.INFO,
        "\t\t\t\t\t\t... Process Duration:"
        "{value:1.{precision}f} s".format(
            value=time.time() - start_time, precision=precision
        ),
    )
